(*  Title:      Nitpick/Tools/nitpick.ML
    ID:         $Id: nitpick.ML,v 1.81 2008/10/10 15:03:21 blanchet Exp $
    Author:     Jasmin Blanchette, TU Muenchen
    Copyright   2008, 2009

Finite model generation for HOL formulas using Kodkod.
*)

signature NITPICK =
sig
  type hol_term = Term.term
  datatype mood = Optimistic | Realistic | Pessimistic

  type params = {
    cards_assigns: (string * int list) list,
    mults_assigns: (string * int list) list,
    iters_assigns: (string * int list) list,
    lockstep: bool,
    sat_solver: string,
    batch_size: int,
    falsify: bool,
    normalize: bool,
    inductive_mood: mood,
    specialize_funs: bool,
    skolem_depth: int,
    peephole: bool,
    timeout: Time.time,
    auto_timeout: Time.time,
    wellfounded_timeout: Time.time,
    sym_break: int,
    sharing_depth: int,
    flatten_props: bool,
    verbose: bool,
    show_skolems: bool,
    show_datatypes: bool,
    show_consts: bool,
    watch: hol_term list,
    max_potential: int,
    check_potential: bool,
    check_genuine: bool,
    expect: string}

  val satisfy_or_falsify_term : Proof.context -> params -> hol_term -> string
  val satisfy_or_falsify_subgoal :
    Proof.context -> params -> thm -> int -> string
end;

structure Nitpick : NITPICK =
struct

open NitpickUtil
open NitpickHOL
open NitpickScope
open NitpickPeephole
open NitpickRep
open NitpickTerm
open NitpickKodkod
open NitpickModel

type params = {
  cards_assigns: (string * int list) list,
  mults_assigns: (string * int list) list,
  iters_assigns: (string * int list) list,
  lockstep: bool,
  sat_solver: string,
  batch_size: int,
  falsify: bool,
  normalize: bool,
  inductive_mood: mood,
  specialize_funs: bool,
  skolem_depth: int,
  peephole: bool,
  timeout: Time.time,
  auto_timeout: Time.time,
  wellfounded_timeout: Time.time,
  sym_break: int,
  sharing_depth: int,
  flatten_props: bool,
  verbose: bool,
  show_skolems: bool,
  show_datatypes: bool,
  show_consts: bool,
  watch: hol_term list,
  max_potential: int,
  check_potential: bool,
  check_genuine: bool,
  expect: string}

(* ## have reconstruct_hol_model take this data structure *)
type problem_extension = {
  free_names: term list,
  ctor_names: term list,
  non_ctor_names: term list,
  rel_table: term NameTable.table,
  axioms: Kodkod.formula list,
  core: Kodkod.formula}

type rich_problem = Kodkod.problem * problem_extension

(* *)
fun satisfy_or_falsify_term ctx params orig_t =
  let
    val timer = Timer.startRealTimer ()
    val _ = priority "Nitpicking\\<dots>"
(*
    val _ = tracing ("*** " ^ Syntax.string_of_term ctx orig_t)
*)
    val {cards_assigns, mults_assigns, iters_assigns, lockstep, sat_solver,
         batch_size, falsify, normalize, inductive_mood, specialize_funs,
         skolem_depth, peephole, timeout, auto_timeout, wellfounded_timeout,
         sym_break, sharing_depth, flatten_props, verbose, show_skolems,
         show_datatypes, show_consts, watch, max_potential, check_potential,
         check_genuine, expect} = params
    val _ =
      if exists (not o member op= (Term.add_frees orig_t []))
                (fold Term.add_frees watch []) then
        error ("The \"watch\" parameter specifies unknown free variables.")
      else
        ()
    val original_max_potential = max_potential
    val thy = ProofContext.theory_of ctx
    val simp_table = ref (const_simp_table ctx)
    val psimp_table = ref (const_psimp_table ctx)
    val intro_table = ref (inductive_intro_table ctx)
    val (ext_ctx as {skolems, fun_specials, ...}) =
      {thy = thy, ctx = ctx, inductive_mood = inductive_mood,
       specialize_funs = specialize_funs, skolem_depth = skolem_depth,
       wellfounded_timeout = wellfounded_timeout, verbose = verbose,
       watch = watch, simp_table = simp_table, psimp_table = psimp_table,
       intro_table = intro_table, skolems = ref [], fun_specials = ref [],
       wfs = ref []}
    val _ = null (Term.add_tvars orig_t [])
            orelse raise NotSupported ("satisfy_or_falsify_term",
                                       "schematic type variables")
    val vars = sort_wrt (fst o fst) (Term.add_vars orig_t [])
    val norm_t = (normalize ? Nbe.norm_term thy) orig_t handle _ => orig_t
    val _ = if norm_t <> orig_t then
              priority ("Normalized formula: "
                        ^ quote (Syntax.string_of_term ctx norm_t) ^ ".")
            else
              ()
    val ex_closed_t =
      fold (fn ((s, j), T) => fn t =>
               HOLogic.exists_const T
               $ Abs (s, T, abstract_over (Var ((s, j), T), t)))
           vars norm_t
    val (axiom_ts, core_t) =
      ((normalize ? Nbe.norm_term thy) ex_closed_t handle _ => ex_closed_t)
      |> preprocess_hol_term ext_ctx falsify
    val Ts = ground_types_in_hol_term ctx (core_t :: axiom_ts)
(*
    val _ = tracing "Types:"
    val _ = map (tracing o Syntax.string_of_typ ctx) Ts
    val _ = tracing "Axioms:"
    val _ = map (tracing o Syntax.string_of_term ctx) axiom_ts
    val _ = tracing "Core:"
    val _ = tracing (Syntax.string_of_term ctx core_t)
*)
    val axiom_us = map (from_hol_term thy StrictEq) axiom_ts
    val core_u = from_hol_term thy Eq core_t
    val fake_u = Tuple (bool_T, Any, core_u :: axiom_us) (* ## cleanup *)
    val free_names = free_names fake_u
    val const_names = const_names fake_u
    val non_ctor_names = filter (not o is_ctor thy) const_names
    val dodgy_ind_pred_names =
      if inductive_mood = Optimistic then
        filter (fn ConstName (s, T, _) => is_inductive_pred thy (s, T)
                 | _ => false) non_ctor_names
      else
        []
(*
    val _ = map (tracing o term_to_string ctx) axiom_us
    val _ = tracing (term_to_string ctx core_u)
*)

    (* *)
    val kodkod_sat_solver =
      let
        val name =
          if sat_solver <> "auto" then sat_solver
          else if getenv "JERUSAT_HOME" <> "" then "Jerusat"
          else if getenv "BERKMIN_HOME" <> "" then "BerkMin"
          else if getenv "BERKMINALLOY_HOME" <> "" then "BerkMinAlloy"
          else if getenv "RSAT_HOME" <> "" then "RSat"
          else if getenv "ZCHAFF_HOME" <> "" then "zChaff"
          else if getenv "MINISAT_HOME" <> "" then "MiniSat"
          else "SAT4J"
        (* bool -> bool -> string -> string -> string * list -> string list
           -> string list *)
        fun aux kodkod_friendly output_to_file dir_var executable args markers =
          case getenv dir_var of
            "" =>
            if kodkod_friendly then [name]
            else error ("SAT solver " ^ quote name ^ " is not configured.")
          | dir =>
            let
              val temp = getenv "ISABELLE_TMP" ^ "/" ^ name ^ serial_string ()
              val cnf_file = temp ^ ".cnf"
              val out_file = temp ^ ".out"
            in
              [if null markers then "External" else "ExternalV2",
               dir ^ "/" ^ executable, cnf_file,
               if output_to_file then out_file else ""] @ markers @
              (if output_to_file then [out_file] else []) @ args
            end
        val _ = if verbose andalso sat_solver = "auto" then
                  priority ("Using SAT solver " ^ quote name ^ ".")
                else
                  ()
      in
        case name of
          "SAT4J" => ["DefaultSAT4J"]
        | "SAT4JLight" => ["LightSAT4J"]
        | "Jerusat" => aux false false "JERUSAT_HOME" "Jerusat1.3" [] []
        | "BerkMin" =>
          let val exec = getenv "BERKMIN_EXE" in
            aux false false "BERKMIN_HOME"
                (if exec = "" then "BerkMin561" else exec) []
                ["Satisfiable          !!", "solution =",
                 "UNSATISFIABLE          !!"]
          end
        | "BerkMinAlloy" => aux false false "BERKMINALLOY_HOME" "berkmin" [] []
        | "RSat" => aux false false "RSAT_HOME" "rsat" ["-s"]
                        ["s SATISFIABLE", "v ", "s UNSATISFIABLE"]
        | "zChaff" => aux true false "ZCHAFF_HOME" "zchaff" []
                          ["Instance Satisfiable", "", "Instance Unsatisfiable"]
        | "MiniSat" => aux true true "MINISAT_HOME" "minisat" []
                           ["SAT", "", "UNSAT"]
        | _ => error ("Unknown SAT solver " ^ quote name ^ ".")
      end

    (* ## move to nitpick_kodkod.ML *)
    (* bool -> scope -> rich_problem option *)
    fun problem_for_scope max_potential scope =
      let
(*
        val _ = tracing (string_for_scope "; " scope)
*)
        val dtypes = #datatypes scope
        val nat_card = card_of_type (#card_assigns scope) nat_T
        val int_card = card_of_type (#card_assigns scope) int_T
        val kk = kodkod_ctors peephole nat_card int_card
        val (free_names, rep_table) =
          choose_reps_for_free_vars scope free_names NameTable.empty
        val (ctor_names, rep_table) = choose_reps_for_all_ctors scope rep_table
        val (non_ctor_names, rep_table) =
          choose_reps_for_consts scope non_ctor_names rep_table
        val axiom_us =
          map (choose_reps_in_term true Formula scope rep_table) axiom_us
        val core_u = choose_reps_in_term false Formula scope rep_table core_u
(*
        val _ = map (fn x => tracing (term_to_string ctx x))
                    (free_names @ const_names)
        val _ = map (tracing o term_to_string ctx) axiom_us
        val _ = tracing (term_to_string ctx core_u)
*)
        val (free_rels, pool, rel_table) =
          rename_free_vars free_names initial_pool NameTable.empty
        val (ctor_rels, pool, rel_table) =
          rename_free_vars ctor_names pool rel_table
        val (non_ctor_rels, pool, rel_table) =
          rename_free_vars non_ctor_names pool rel_table
        val core_u = rename_vars_in_term pool rel_table core_u
        val axiom_us = map (rename_vars_in_term pool rel_table) axiom_us
(*
        val _ = map (fn x => tracing (term_to_string ctx x))
                    (free_rels @ ctor_rels @ non_ctor_rels)
        val _ = map (tracing o term_to_string ctx) axiom_us
        val _ = tracing (term_to_string ctx core_u)
*)
        val axiom_fs =
          map (to_kodkod_formula dtypes peephole false true false kk) axiom_us
        val core_f = to_kodkod_formula dtypes peephole (max_potential > 0)
                                       (max_potential > 0) falsify kk core_u
        val formula = fold s_and axiom_fs core_f
        val comment = spacious_string_for_scope scope
        val settings = [("solver", commas (map quote kodkod_sat_solver)),
                        ("skolem_depth", "-1"),
                        ("bit_width", "16"),
                        ("symmetry_breaking", Int.toString sym_break),
                        ("sharing", Int.toString sharing_depth),
                        ("flatten", Bool.toString flatten_props)]
        val plain_rels = free_rels @ non_ctor_rels
        val plain_bounds = map bound_for_plain_rel plain_rels
        val plain_axioms = map (axiom_for_plain_rel peephole kk) plain_rels
        val ctor_bounds = map (bound_for_ctor_rel dtypes) ctor_rels
        val ctor_axioms = map (axiom_for_ctor_rel peephole dtypes kk) ctor_rels
        val nfa_table = map (nfa_entry_for_dtype kk dtypes rel_table) dtypes
        val dtype_axioms = map (acyclicity_axiom kk nfa_table o #typ) dtypes
        val low_level_axioms = plain_axioms @ ctor_axioms @ dtype_axioms
        val univ_card =
          univ_card nat_card int_card (plain_bounds @ ctor_bounds) formula
        val built_in_bounds =
          bounds_for_built_in_rels_in_formula (max_potential > 0) univ_card
                                              nat_card int_card formula
        val bounds =
          coalesce_bounds (built_in_bounds @ plain_bounds @ ctor_bounds) []
        val formula = fold_rev s_and low_level_axioms formula
      in
        SOME ({comment = comment, settings = settings, univ_card = univ_card,
               tuple_assigns = [], bounds = bounds, int_bounds = [],
               expr_assigns = [], formula = formula},
              {free_names = free_names, ctor_names = ctor_names,
               non_ctor_names = non_ctor_names, rel_table = rel_table,
               axioms = axiom_fs @ low_level_axioms, core = core_f})
      end
      handle Limit _ => NONE

    (* *)
    fun tuple_set_for_rel univ_card =
      Kodkod.TupleSet o map (kk_tuple univ_card) o the oo AList.lookup op=
    (* *)
    fun debase_simple_bound univ_card [] raw_bounds _ (x, s) =
        ([(x, s)], [tuple_set_for_rel univ_card raw_bounds x])
      | debase_simple_bound univ_card (r :: rs) raw_bounds tss (x, s) =
        (case r of
           FreeRel (x', _, R, _) =>
           if x = x' then
             ([(x, s)],
              if body_rep R = Formula then
                [Kodkod.TupleSet [], tuple_set_for_rel univ_card raw_bounds x]
              else
                tss)
           else
             debase_simple_bound univ_card rs raw_bounds tss (x, s)
         | _ => raise BadTerm ("debase_simple_bound", [r]))
    (* *)
    fun debase_bound univ_card dodgy_rels raw_bounds (zs, tss) =
      map (debase_simple_bound univ_card dodgy_rels raw_bounds tss) zs

    (* rich_problem -> Kodkod.raw_bound list -> Kodkod.problem *)
    fun debase_inductive_problem ({comment, settings, univ_card, tuple_assigns,
                                   bounds, int_bounds, expr_assigns, ...},
                                  {rel_table, axioms, core, ...}) raw_bounds =
      let val dodgy_rels = map (lookup_name rel_table) dodgy_ind_pred_names in
        {comment = "debased\n" ^ comment, settings = settings,
         univ_card = univ_card, tuple_assigns = tuple_assigns,
         bounds = List.concat (map (debase_bound univ_card dodgy_rels
                                                 raw_bounds) bounds),
         int_bounds = int_bounds, expr_assigns = expr_assigns,
         formula = fold s_and axioms (Kodkod.Not core)}
      end

    val the_word_model = if falsify then "counterexample" else "model"

    datatype model_kind = Genuine | Potential | GenuineModIndPreds

    (* scope -> problem_extension -> model_kind -> raw_bound list
       -> bool option *)
    fun print_and_check_model scope
            {free_names, ctor_names, non_ctor_names, rel_table, ...}
            kind bounds =
      (priority
           ("Nitpick found a " ^
            (if kind = Genuine then "genuine" else "potential") ^ " " ^
            the_word_model ^
            (if is_empty_scope scope then ""
             else " for " ^ compact_string_for_scope scope) ^
            ":\n\n" ^
            (reconstruct_hol_model
                 {show_skolems = show_skolems, show_datatypes = show_datatypes,
                  show_consts = show_consts}
                 scope watch (!skolems) (!fun_specials) free_names ctor_names
                 non_ctor_names rel_table bounds
             |> prefix_lines "    "));
       if kind = Genuine then
         if check_genuine then
           ((case prove_hol_model scope auto_timeout free_names ctor_names
                                  rel_table bounds falsify orig_t of
              SOME true => priority ("Confirmation by \"auto\": The above " ^
                                     the_word_model ^ " is really genuine.")
            | SOME false =>
              error ("A supposedly genuine " ^ the_word_model ^ " was shown to \
                     \be spurious by \"auto\".\nThis should never happen.\n\
                     \Please send a bug report to blanchette\
                     \@in.tum.de.")
            | NONE => priority "No confirmation by \"auto\"."); NONE)
         else
           NONE
       else
         if kind = Potential andalso check_potential then
           let
             val status = prove_hol_model scope auto_timeout free_names
                                          ctor_names rel_table bounds falsify
                                          orig_t
           in
             (case status of
                SOME true => priority ("Confirmation by \"auto\": The above " ^
                                       the_word_model ^ " is genuine.")
              | SOME false => priority ("Refutation by \"auto\": The above " ^
                                        the_word_model ^ " is spurious.")
              | NONE => priority "No confirmation by \"auto\".");
             status
           end
         else
           NONE)
    (* instead of NONE and SOME, use Stop and Continue *)
    (* int * bool -> scope list -> Kodkod.problem list
       -> (int * bool) option *)
    fun solve_any_problem (max_potential, donno) scopes problems =
      let
        (* *)
        fun print_and_check j =
          print_and_check_model (nth scopes j) (snd (nth problems j))
        (* *)
        fun is_problematic_inductive_pred j raw_bounds1 raw_bounds2 name =
          let val {rel_table, ...} = snd (nth problems j) in
            tuple_list_for_name rel_table raw_bounds1 name
            <> tuple_list_for_name rel_table raw_bounds2 name
          end
        (* *)
        fun handle_problematic_inductive_preds names =
          let
            val ss = sort_distinct string_ord
                         (map (quote o original_name o nickname_of) names)
          in
            priority ((if length names = 1 then
                         "The inductive predicate " ^ hd ss
                       else
                         "The inductive predicates " ^ serial_commas ss) ^
                      " might be non-wellfounded. Bailing out. Try again with \
                      \\"inductive_mood\" set to \"realistic\" or \
                      \\"pessimistic\"."); NONE
          end
        (* *)
        fun handle_genuine_model_mod_inductive_preds j bounds =
          if null dodgy_ind_pred_names then
            (print_and_check j Genuine bounds; NONE)
          else
            (if max_potential > 0 then
               print_and_check j GenuineModIndPreds bounds
             else
               NONE;
             case Kodkod.solve_any_problem timeout
                      [debase_inductive_problem (nth problems j) bounds] of
               Kodkod.Satisfiable (_, new_bounds) =>
               handle_problematic_inductive_preds
                   (filter (is_problematic_inductive_pred j bounds new_bounds)
                           dodgy_ind_pred_names)
             | Kodkod.Unsatisfiable =>
               if max_potential > 0 then
                 (priority ("Confirmation by Kodkod: The above " ^
                            the_word_model ^ " is genuine."); NONE)
               else
                 (print_and_check j Genuine bounds; NONE)
             | _ => handle_problematic_inductive_preds dodgy_ind_pred_names)
      in
        case Kodkod.solve_any_problem timeout (map fst problems) of
          Kodkod.Satisfiable (j, bounds) =>
          if is_only_liberally_satisfiable bounds then
            if print_and_check j Potential bounds = SOME true then
              NONE
            else
              solve_any_problem (max_potential - 1, donno)
                  (List.drop (scopes, j))
                  (if max_potential > 1 then
                     apfst illiberalize_problem (nth problems j)
                     :: List.drop (problems, j + 1)
                   else
                     map (apfst illiberalize_problem) (List.drop (problems, j)))
          else
            handle_genuine_model_mod_inductive_preds j bounds
        | Kodkod.Unsatisfiable => SOME (max_potential, donno)
        | Kodkod.Error s =>
          (if verbose orelse not donno then warning ("Kodkod error: " ^ s ^ ".")
           else ();
           SOME (max_potential, true))
      end

    (* int * bool -> scope list -> (int * bool) option *)
    fun run_batch (max_potential, donno) scopes =
      let
        val _ =
          if verbose then
            priority ("Trying " ^ Int.toString (length scopes) ^ " scope" ^
                      plural_s scopes ^ ":\n" ^
                      space_implode ";\n"
                          (map compact_string_for_scope scopes) ^ ".")
          else
            ()
        val problems = List.mapPartial (problem_for_scope max_potential) scopes
      in
        solve_any_problem (max_potential,
                           donno orelse length problems < length scopes)
                          scopes problems
      end

    (* int * bool -> scope list -> Kodkod.outcome *)
    fun run_batches (max_potential, donno) [] =
        if donno then
          (priority "Nitpick encountered some problem. The search is \
                    \inconclusive."; "unknown")
        else if max_potential = original_max_potential then
          (priority ("Nitpick found no " ^ the_word_model ^ "."); "unsat")
        else
          (priority ("Nitpick could not find a better " ^ the_word_model ^ ".");
           "weaksat")
      | run_batches (max_potential, donno) (batch :: batches) =
        (case run_batch (max_potential, donno) batch of
           SOME (max_potential, donno) =>
           run_batches (max_potential, donno) batches
         | NONE => "sat")

    (* unit -> string *)
    fun run_the_batches () =
      run_batches (max_potential, false)
                  (batch_list batch_size
                       (all_scopes ctx cards_assigns mults_assigns iters_assigns
                                   Ts lockstep))

    val outcome_code =
      time_limit timeout run_the_batches ()
      handle TimeLimit.TimeOut => (priority "Nitpick ran out of time.";
                                   "unknown")
  in
    if expect = "" orelse outcome_code = expect then outcome_code
    else error ("Unexpected outcome: " ^ quote outcome_code ^ ".")
  end

(* Proof.context -> params -> thm -> int -> string *)
fun satisfy_or_falsify_subgoal ctx params thm subgoal =
  satisfy_or_falsify_term ctx params (nth (prems_of thm) subgoal)

end;
