(*  Title:      Nitpick/Tools/nitpick_hol.ML
    ID:         $Id: nitpick_hol.ML,v 1.81 2008/12/19 15:19:47 blanchet Exp $
    Author:     Jasmin Blanchette, TU Muenchen
    Copyright   2008, 2009

Auxiliary HOL-related functions used by Nitpick.
*)

signature NITPICK_HOL =
sig
  type hol_term = Term.term
  type const_table = hol_term list Symtab.table
  type fun_special =
    ((string * typ) * int list * hol_term list) * (string * typ)
  type wf_cache = ((string * typ) * bool) list

  datatype mood = Optimistic | Realistic | Pessimistic

  type extended_context = {
    thy: theory,
    ctx: Proof.context,
    inductive_mood: mood,
    specialize_funs: bool,
    skolem_depth: int,
    wellfounded_timeout: Time.time,
    verbose: bool,
    watch: hol_term list,
    simp_table: const_table ref,
    psimp_table: const_table ref,
    intro_table: const_table ref,
    skolems: (string * string list) list ref,
    fun_specials: fun_special list ref,
    wfs: wf_cache ref}

  val num_prefix : string
  val skolem_prefix : string
  val watch_prefix : string
  val is_TFree : typ -> bool
  val is_fun_type : typ -> bool
  val is_set_type : typ -> bool
  val is_pair_type : typ -> bool
  val is_iterator_type : typ -> bool
  val is_integer_type : typ -> bool
  val second_range_type : typ -> typ
  val third_range_type : typ -> typ
  val original_name : string -> string
  val user_friendly_const :
    hol_term list -> (string * string list) list -> fun_special list
    -> string * typ -> hol_term
  val type_to_string : Proof.context -> typ -> string
  val instantiate_type : theory -> typ -> typ -> typ -> typ
  val is_abs_fun : theory -> string * typ -> bool
  val is_rep_fun : theory -> string * typ -> bool
  val eta_expand : typ list -> hol_term -> int -> hol_term
  val extensionalize : hol_term -> hol_term
  val const_simp_table : Proof.context -> const_table
  val const_psimp_table : Proof.context -> const_table
  val inductive_intro_table : Proof.context -> const_table
  val is_inductive_pred : theory -> string * typ -> bool
  val preprocess_hol_term :
    extended_context -> bool -> hol_term -> hol_term list * hol_term
  val ground_types_in_hol_term : Proof.context -> hol_term list -> typ list
end;

structure NitpickHOL : NITPICK_HOL =
struct

open NitpickUtil

type const_table = hol_term list Symtab.table
type fun_special = ((string * typ) * int list * hol_term list) * (string * typ)
type wf_cache = ((string * typ) * bool) list

datatype mood = Optimistic | Realistic | Pessimistic

type extended_context = {
  thy: theory,
  ctx: Proof.context,
  inductive_mood: mood,
  specialize_funs: bool,
  skolem_depth: int,
  wellfounded_timeout: Time.time,
  verbose: bool,
  watch: hol_term list,
  simp_table: const_table ref,
  psimp_table: const_table ref,
  intro_table: const_table ref,
  skolems: (string * string list) list ref,
  fun_specials: fun_special list ref,
  wfs: wf_cache ref}

val num_prefix = nitpick_prefix ^ "num_"
val it_prefix = nitpick_prefix ^ "it_"
val wf_prefix = nitpick_prefix ^ "wf_"
val skolem_prefix = nitpick_prefix ^ "skolem"
val special_prefix = nitpick_prefix ^ "special"
val watch_prefix = nitpick_prefix ^ "watch"

(* int -> string *)
fun skolem_prefix_for j = skolem_prefix ^ Int.toString j ^ "_"
fun special_prefix_for j = special_prefix ^ Int.toString j ^ "_"

(* typ -> bool *)
fun is_TFree (TFree _) = true
  | is_TFree _ = false
fun is_fun_type (Type ("fun", _)) = true
  | is_fun_type _ = false
fun is_set_type (Type ("fun", [_, Type ("bool", [])])) = true
  | is_set_type _ = false
fun is_pair_type (Type ("*", _)) = true
  | is_pair_type _ = false
fun is_iterator_type (Type (s, _)) = String.isPrefix it_prefix s
  | is_iterator_type _ = false
fun is_integer_type T = (T = nat_T orelse T = int_T orelse is_iterator_type T)

(* typ -> typ *)
val second_range_type = range_type o range_type
val third_range_type = range_type o second_range_type

(* string -> string *)
fun original_name s =
  if String.isPrefix nitpick_prefix s then
    original_name (Substring.string (Substring.triml 1
        (snd (Substring.position "_" (Substring.full s)))))
  else
    s

(* hol_term list -> (string * string list) list -> fun_special list
   -> string * typ -> hol_term *)
fun user_friendly_const watch skolems fun_specials (s, T) =
  if String.isPrefix special_prefix s then
    let
      val ((s', T'), js, ts) = the_single (AList.find op= fun_specials (s, T))
      val max_j = List.last js
      val Ts = List.take (binder_types T', max_j + 1)
      val ts' = map2 (fn j => fn T =>
                         AList.lookup op= (js ~~ ts) j
                         |> the_default (Const ("\\<star>", T)))
                     (0 upto max_j) Ts
    in
      list_comb (user_friendly_const watch skolems fun_specials (s', T'), ts')
    end
  else if String.isPrefix wf_prefix s then
    lambda (Free ("i", nat_T)) (Const (original_name s, range_type T))
  else if String.isPrefix skolem_prefix s then
    let
      val (binder_Ts, body_T) = strip_type T
      val ss = the (AList.lookup op= skolems s)
      val (Ts, Ts') = chop (length ss) binder_Ts
      val frees = map Free (ss ~~ Ts)
    in fold lambda frees (Const (original_name s, Ts' ---> T)) end
  else if String.isPrefix watch_prefix s then
    nth watch (the (Int.fromString (unprefix watch_prefix s)))
  else
    Const (original_name s, T)

(* Proof.context -> typ -> string *)
fun type_to_string ctx T = (* ## ctx isn't used *)
  case T of
    Type (s, Ts) =>
    if is_iterator_type T then
      type_to_string ctx (Type (original_name s, tl Ts))
    else
      (case Ts of
         [] => s
       | [T] => type_to_string ctx T ^ " " ^ s
       | Ts => "(" ^ commas (map (type_to_string ctx) Ts) ^ ") " ^ s)
  | TFree (s, _) => s
  | TVar ((s, _), _) => s

(* *)
val varify_type = map_type_tfree (fn (s, S) => TVar ((s, 0), S))

(* theory -> typ -> typ -> typ -> typ *)
fun instantiate_type thy T1 T1' T2 =
  Envir.typ_subst_TVars (Sign.typ_match thy (varify_type T1, T1') Vartab.empty)
                        (varify_type T2)
  handle Type.TYPE_MATCH => raise BadType ("instantiate_type", [T1, T1'])

(* theory -> string * typ -> bool *)
fun is_abs_fun thy (s, Type ("fun", [_, Type (s', _)])) =
    (case TypedefPackage.get_info thy s' of
       SOME {Abs_name, ...} => (s = Abs_name)
     | NONE => false)
  | is_abs_fun _ _ = false
fun is_rep_fun thy (s, Type ("fun", [Type (s', _), _])) =
    (case TypedefPackage.get_info thy s' of
       SOME {Rep_name, ...} => (s = Rep_name)
     | NONE => false)
  | is_rep_fun _ _ = false

(* typ list -> hol_term -> int -> hol_term *)
fun eta_expand _ t 0 = t
  | eta_expand Ts t n =
    fold_rev (curry3 Abs ("x\\<^isub>\\<eta>" ^ nat_subscript n))
             (List.take (binder_types (fastype_of1 (Ts, t)), n))
             (list_comb (incr_boundvars n t, map Bound (n - 1 downto 0)))

(* hol_term -> hol_term *)
fun extensionalize t =
  case t of
    @{const Trueprop} $ t1 => extensionalize t1
  | Const (@{const_name "op ="}, _) $ t1 $ Abs (s, T, t2) =>
    let val v = Var ((s, maxidx_of_term t + 1), T) in
      extensionalize (HOLogic.mk_eq (t1 $ v, subst_bound (v, t2)))
    end
  | _ => t

(* hol_term -> hol_term list -> hol_term *)
val betapplys_on = foldl (betapply o swap)

val built_in_consts =
  [(@{const_name all}, 1),
   (@{const_name "=="}, 2),
   (@{const_name "==>"}, 2),
   (@{const_name Trueprop}, 1),
   (@{const_name Not}, 1),
   (@{const_name False}, 0),
   (@{const_name True}, 0),
   (@{const_name All}, 1),
   (@{const_name Ex}, 1),
   (@{const_name "op ="}, 2),
   (@{const_name "op &"}, 2),
   (@{const_name "op |"}, 2),
   (@{const_name "op -->"}, 2),
   (@{const_name If}, 3),
   (@{const_name Let}, 2),
   (@{const_name Unity}, 0),
   (@{const_name Pair}, 2),
   (@{const_name fst}, 1),
   (@{const_name snd}, 1),
   (@{const_name Id}, 0),
   (@{const_name "{}"}, 0),
   (@{const_name UNIV}, 0),
   (@{const_name "op :"}, 2),
   (@{const_name insert}, 2),
   (@{const_name trancl}, 1),
   (@{const_name rtrancl}, 1),
   (@{const_name "op Un"}, 2),
   (@{const_name "op Int"}, 2),
   (@{const_name Suc}, 0),
   (@{const_name finite}, 1),
   (@{const_name card}, 1),
   (@{const_name nat}, 0)]
val built_in_typed_consts =
  [((@{const_name uminus_class.uminus}, int_T --> int_T), 0),
   ((@{const_name of_nat}, nat_T --> int_T), 0)]
val built_in_set_consts =
  [(@{const_name minus_class.minus}, 2),
   (@{const_name ord_class.less}, 2),
   (@{const_name ord_class.less_eq}, 2),
   (@{const_name uminus_class.uminus}, 2)]
val built_in_integer_consts =
  [(@{const_name zero_class.zero}, 0),
   (@{const_name one_class.one}, 0),
   (@{const_name plus_class.plus}, 0),
   (@{const_name minus_class.minus}, 0),
   (@{const_name times_class.times}, 0),
   (@{const_name ord_class.less}, 2),
   (@{const_name ord_class.less_eq}, 2)]
val unsupported_consts =
  [@{const_name Abs_unit},
   @{const_name Rep_unit},
   @{const_name Abs_Prod},
   @{const_name Rep_Prod},
   @{const_name Abs_Sum},
   @{const_name Rep_Sum},
   @{const_name Abs_Integ},
   @{const_name Rep_Integ}]

(* string * typ -> int option *)
fun arity_of_built_in_const (s, T) =
  case AList.lookup op= built_in_consts s of
    SOME n => SOME n
  | NONE =>
    case AList.lookup op= built_in_typed_consts (s, T) of
      SOME n => SOME n
    | NONE =>
      if is_fun_type T andalso is_set_type (domain_type T) then
        AList.lookup op= built_in_set_consts s
      else if is_integer_type T
              orelse (is_fun_type T
                      andalso is_integer_type (domain_type T)) then
        AList.lookup op= built_in_integer_consts s
      else
        NONE
(* string * typ -> bool *)
val is_built_in_const = isSome o arity_of_built_in_const

(* *)
fun const_under_def t =
  case t of
    @{const Trueprop} $ t1 => const_under_def t1
  | @{const "==>"} $ _ $ t2 => const_under_def t2
  | Const (@{const_name "=="}, _) $ t1 $ _ => const_under_def t1
  | Const (@{const_name "op ="}, _) $ t1 $ _ => const_under_def t1
  | _ => case head_of t of
           Const x => SOME x
         | _ => NONE

(* hol_term -> (string * hol_term) option *)
fun pair_for_prop prop =
  case const_under_def prop of
    SOME x => if is_built_in_const x then NONE else SOME (fst x, prop)
  | _ => raise BadHOLTerm ("pair_for_prop", [prop])

(* *)
val table_for_theorems =
  Symtab.make o AList.group op= o List.mapPartial (pair_for_prop o Thm.prop_of)

(* Proof.context -> const_table *)
val const_simp_table = Nitpick_Const_Simp_Thms.get #> table_for_theorems
val const_psimp_table = Nitpick_Const_Psimp_Thms.get #> table_for_theorems
val inductive_intro_table = Nitpick_Ind_Intro_Thms.get #> table_for_theorems

(* *)
fun props_for_const thy table (s, T) =
  case Symtab.lookup table s of
    SOME ts =>
    List.mapPartial (fn t => SOME (Refute.specialize_type thy (s, T) t)
                             handle Type.TYPE_MATCH => NONE) ts
  | NONE => []

(* string -> string -> string *)
fun prefix_with_const_name s = prefix (List.last (space_explode "." s) ^ ".")
(* string -> hol_term -> hol_term *)
val prefix_abs_vars_with_const_name = Term.map_abs_vars o prefix_with_const_name

(* *)
fun def_of_const thy (s, T) =
  case Refute.get_def thy (s, T) of
    SOME (_, def) => SOME (prefix_abs_vars_with_const_name s def)
  | NONE => NONE

datatype fixpoint_kind = Lfp | Gfp | NoFp

(* hol_term -> fixpoint_kind *)
fun fixpoint_kind_of_rhs (Abs (_, _, t)) = fixpoint_kind_of_rhs t
  | fixpoint_kind_of_rhs (Const (@{const_name lfp}, _) $ Abs _) = Lfp
  | fixpoint_kind_of_rhs (Const (@{const_name gfp}, _) $ Abs _) = Gfp
  | fixpoint_kind_of_rhs _ = NoFp
(* theory -> string * typ -> fixpoint_kind *)
fun fixpoint_kind_of_const thy x =
  if is_built_in_const x then NoFp
  else fixpoint_kind_of_rhs (the (def_of_const thy x)) handle Option => NoFp
(* theory -> (string * typ) -> bool *)
val is_inductive_pred = equal Lfp oo fixpoint_kind_of_const
(* extended_context -> (string * typ) -> bool *)
fun is_equational_fun (ext_ctx as {thy, simp_table, psimp_table, ...}) x =
  is_inductive_pred thy x
  orelse not (null (props_for_const thy (!simp_table) x))
  orelse not (null (props_for_const thy (!psimp_table) x))

exception NoTriple

(* *)
fun triple_for_intro_rule thy x t =
  let
    val prems = Logic.strip_imp_prems t |> map (ObjectLogic.atomize_term thy)
    val concl = Logic.strip_imp_concl t |> ObjectLogic.atomize_term thy
    val (main, side) = List.partition (fn t => x mem Term.add_consts t []) prems
    (* hol_term -> bool *)
    val good_head = equal (Const x) o head_of
  in if forall good_head main then (side, main, concl) else raise NoTriple end

(* *)
fun tuple_for_call t = foldl1 HOLogic.mk_prod (snd (strip_comb t))

(* *)
fun wf_constraint_for rel side concl main =
  let
    val core = HOLogic.mk_mem (HOLogic.mk_prod (tuple_for_call main,
                                                tuple_for_call concl), Var rel)
    val t = foldl HOLogic.mk_imp core side
    val vars = filter (not_equal rel) (Term.add_vars t [])
  in
    Library.foldl (fn (t', ((x, i), T)) =>
                      HOLogic.all_const T
                      $ Abs (x, T, abstract_over (Var ((x, i), T), t')))
                  (t, vars)
  end

(* *)
fun wf_constraint_for_triple rel (side, main, concl) =
  map (wf_constraint_for rel side concl) main |> foldl1 HOLogic.mk_conj

(* *)
fun terminates_by ctx timeout goal tac =
  (goal |> SINGLE (CLASET Classical.safe_tac) |> the
        |> SINGLE (timeout_tac timeout
                               (tac ctx (auto_tac (local_clasimpset_of ctx))))
        |> the |> Goal.finish; true)
  handle _ => false

(* *)
fun is_wellfounded_inductive_pred_no_cache ctx timeout intro_tab (s, T) =
  let
    val thy = ProofContext.theory_of ctx
    val ts = props_for_const thy intro_tab (s, T)
    val triples = map (triple_for_intro_rule thy (s, T)) ts
                  |> filter (not o null o #2)
  in
    if null triples then
      true
    else
      let
        val binders_T = foldl1 HOLogic.mk_prodT (binder_types T)
        val rel_T = HOLogic.mk_prodT (binders_T, binders_T) --> bool_T
        val j = foldl Int.max 0 (map maxidx_of_term ts) + 1
        val rel = (("R", j), rel_T)
        val prop = Const (@{const_name wf}, rel_T --> bool_T) $ Var rel
                   :: map (wf_constraint_for_triple rel) triples
                   |> foldl1 HOLogic.mk_conj |> HOLogic.mk_Trueprop
(*
        val _ = tracing ("***** Goal: " ^ Syntax.string_of_term ctx prop)
*)
        val goal = cterm_of thy prop |> Goal.init
        val tacs = [LexicographicOrder.lexicographic_order_tac,
                    ScnpReconstruct.sizechange_tac]
        val res = silence (exists (terminates_by ctx timeout goal)) tacs
      in res end
  end
  handle Empty => false
       | NoTriple => false

(* *)
fun is_wellfounded_inductive_pred
       {ctx, inductive_mood, wellfounded_timeout, verbose, intro_table, wfs,
        ...} x =
  case inductive_mood of
    Optimistic => true
  | Pessimistic => false
  | Realistic =>
    case AList.lookup op= (!wfs) x of
      SOME wf => wf
    | NONE =>
      let
        val wf = is_wellfounded_inductive_pred_no_cache ctx wellfounded_timeout
                                                        (!intro_table) x
        val _ = if verbose then
                  priority ("The inductively defined predicate " ^
                            quote (Syntax.string_of_term ctx (Const x)) ^ " " ^
                            (if wf then
                               "was proved wellfounded. Nitpick can compute it \
                               \efficiently."
                             else
                               "could not be proved wellfounded. Nitpick will \
                               \unroll the predicate."))
                else
                  ()
        val _ = change wfs (cons (x, wf))
      in wf end

(* *)
fun better_founded_inductive_pred_const (ext_ctx as {thy, simp_table, ...})
                                        (s, T) =
  let
    val iter_T = Type (it_prefix ^ s, [domain_type T])
    val x' as (s', _) = (wf_prefix ^ s, iter_T --> T)
    val surrogate = Const x' $ Const (@{const_name zero_class.zero}, iter_T)
  in
    (if not (is_equational_fun ext_ctx x') then
       let
         val def = the (def_of_const thy (s, T))
         val j = maxidx_of_term def + 1
         val (outer, fp_app) = strip_abs def
         val outer_bounds = map Bound (length outer - 1 downto 0)
         val cur = Var (("k", j + 2), iter_T)
         val next = Const (@{const_name Suc}, iter_T --> iter_T) $ cur
         val rhs =
           case fp_app of
             Const _ $ t =>
             betapply (t, list_comb (Const x', next :: outer_bounds))
           | _ => raise BadHOLTerm ("inductive_fun_axiom", [fp_app])
         val (inner, naked_rhs) = strip_abs rhs
         val all = outer @ inner
         val bounds = map Bound (length all - 1 downto 0)
         val vars = map (fn (s, T) => Var ((s, j), T)) all
         val eq = HOLogic.mk_eq (list_comb (Const x', cur :: bounds), naked_rhs)
                  |> curry subst_bounds (rev vars)
       in
         change simp_table (Symtab.update_new (s', [eq]))
       end
     else
       ());
    surrogate
  end

(* *)
fun raw_inductive_prod_axiom thy x =
  let
    val def = the (def_of_const thy x)
    val (outer, fp_app) = strip_abs def
    val outer_bounds = map Bound (length outer - 1 downto 0)
    val rhs = case fp_app of
                Const _ $ t => betapply (t, list_comb (Const x, outer_bounds))
              | _ => raise BadHOLTerm ("inductive_fun_axiom", [fp_app])
    val (inner, naked_rhs) = strip_abs rhs
    val all = outer @ inner
    val bounds = map Bound (length all - 1 downto 0)
    val j = maxidx_of_term def + 1
    val vars = map (fn (s, T) => Var ((s, j), T)) all
    val eq = HOLogic.mk_eq (list_comb (Const x, bounds), naked_rhs)
  in subst_bounds (rev vars, eq) end

(* extended_context -> (string * typ) -> hol_term list *)
fun raw_equational_fun_axioms
        (ext_ctx as {thy, simp_table, psimp_table, ...}) x =
  case props_for_const thy (!simp_table) x of
    [] => (case props_for_const thy (!psimp_table) x of
             [] => [raw_inductive_prod_axiom thy x]
           | psimps => psimps)
  | simps => simps

val equational_fun_axioms = map extensionalize oo raw_equational_fun_axioms

(* string * typ -> int list *)
fun dead_args_in_term x t =
  let
    (* *)
    fun fun_calls accum (Abs (_, _, t)) _ = fun_calls accum t []
      | fun_calls accum (t1 $ t2) arg_ts =
        fun_calls (fun_calls accum t2 []) t1 (t2 :: arg_ts)
      | fun_calls accum t arg_ts = (t = Const x ? cons arg_ts) accum
    (* *)
    fun call_sets [] [] zs = [zs]
      | call_sets [] yss zs = zs :: call_sets yss [] []
      | call_sets ([] :: _) _ _ = []
      | call_sets ((x :: xs) :: xss) yss zs =
        call_sets xss (xs :: yss) (OrdList.insert TermOrd.term_ord x zs)
    val calls = fun_calls [] t []
    val sets = call_sets calls [] []
  in
    fold_rev (fn (j, set) => (length set = 1 andalso is_Var (hd set)) ? cons j)
             ((0 upto length sets - 1) ~~ sets) []
  end
(* string * typ -> hol_term list -> int list *)
fun dead_args_in_terms x ts =
  fold1 (OrdList.inter int_ord) (map (dead_args_in_term x) ts)

(* *)
fun params_in_equation (@{const Trueprop} $ t1) = params_in_equation t1
  | params_in_equation (@{const "==>"} $ _ $ t1) = params_in_equation t1
  | params_in_equation (Const (s, _) $ t1 $ _) =
    if s = @{const_name "op ="} then snd (strip_comb t1) else []
  | params_in_equation _ = []

(* *)
fun specialize_fun_axiom x js fixed_args x' t =
  let
    val fixed_params = filter_indices js (params_in_equation t)
    fun aux args (Abs (s, T, t)) = list_comb (Abs (s, T, aux [] t), args)
      | aux args (t1 $ t2) = aux (aux [] t2 :: args) t1
      | aux args t =
        if t = Const x then
          list_comb (Const x', filter_out_indices js args)
        else
          let val j = find_index (equal t) fixed_params in
            list_comb (if j >= 0 then nth fixed_args j else t, args)
          end
  in aux [] t end

(* *)
fun specialize_consts_in_hol_term
        (ext_ctx as {specialize_funs, simp_table, fun_specials, ...}) axiom t =
  let
    val black_list = if axiom then the_list (const_under_def t) else []
    (* *)
    fun aux args (Const (x as (s, T))) =
        ((if not (x mem black_list) andalso not (null args)
             andalso is_equational_fun ext_ctx x then
            let
              val candidate_args =
                fold_rev (fn (j, t) => not (loose_bvar (t, 0)) ? cons j)
                         ((0 upto length args - 1) ~~ args) []
              val _ = not (null candidate_args) orelse raise Same
              val old_axs = equational_fun_axioms ext_ctx x
              val dead_params = dead_args_in_terms x old_axs
              val js = OrdList.inter int_ord candidate_args dead_params
              val _ = not (null js) orelse raise Same
              val fixed_args = filter_indices js args
              val extra_args = filter_out_indices js args
            in
              case AList.lookup op= (!fun_specials) (x, js, fixed_args) of
                SOME x' => list_comb (Const x', extra_args)
              | NONE =>
                let
                  val x' =
                    (special_prefix_for (length (!fun_specials) + 1) ^ s,
                     filter_out_indices js (binder_types T) ---> body_type T)
                  val new_axs =
                    map (specialize_fun_axiom x js fixed_args x') old_axs
                  val _ = change fun_specials (cons ((x, js, fixed_args), x'))
                  val _ =
                    change simp_table (Symtab.update_new (fst x', new_axs))
                in list_comb (Const x', extra_args) end
            end
          else
            raise Same)
         handle Same => list_comb (Const x, args))
      | aux args (Abs (s, T, t)) = list_comb (Abs (s, T, aux [] t), args)
      | aux args (t1 $ t2) = aux (aux [] t2 :: args) t1
      | aux args t = list_comb (t, args)
  in (specialize_funs ? aux []) t end

(* *)
fun unfold_defs_in_hol_term (ext_ctx as {thy, ...}) t =
  let
    (* typ list -> hol_term -> hol_term *)
    fun aux Ts t =
      case t of
        (t0 as Const (s as @{const_name Int.number_class.number_of},
                      T as Type ("fun", [_, T']))) $ t1 =>
        if is_integer_type T' then
          Const (num_prefix ^ Int.toString (HOLogic.dest_numeral t1), T')
          handle TERM _ => aux Ts (the (def_of_const thy (s, T)) $ t1)
        else
          betapply (t0, aux Ts t1)
      | Const x $ t1 $ t2 $ t3 => aux_const Ts x [t1, t2, t3]
      | Const x $ t1 $ t2 => aux_const Ts x [t1, t2]
      | Const x $ t1 => aux_const Ts x [t1]
      | Const x => aux_const Ts x []
      | Free _ => t
      | Var _ => t
      | Bound _ => t
      | Abs (s, T, body) => Abs (s, T, aux (T :: Ts) body)
      | t1 $ t2 => betapply (aux Ts t1, aux Ts t2)
    (* typ list -> string * typ -> hol_term list -> hol_term *)
    and aux_const Ts x ts =
      let
        val (arity, const) =
          (case arity_of_built_in_const x of
            SOME n => (n, Const x)
          | NONE =>
            if is_inductive_pred thy x
               andalso not (is_wellfounded_inductive_pred ext_ctx x) then
              (0, better_founded_inductive_pred_const ext_ctx x)
            else if Refute.is_IDT_constructor thy x
                    orelse is_equational_fun ext_ctx x then
              (0, Const x)
            else
              case def_of_const thy x of
                SOME def => (0, aux Ts def)
              | NONE => (0, Const x))
        val const_of_ts = betapplys_on const (map (aux Ts) ts)
                          |> Envir.beta_norm
      in eta_expand Ts const_of_ts (arity nat_minus length ts) end
  in aux [] t end

(* *)
fun watch_axiom_for_hol_term j t =
  HOLogic.mk_eq (Const (watch_prefix ^ Int.toString j, fastype_of t), t)

(* *)
fun axioms_for_hol_term (ext_ctx as {thy, watch, ...}) t =
  let
    val all_axioms = Theory.all_axioms_of thy
    (* The add_axioms_for_term parameter in add_axiom is ludicrous but necessary
       to work around a Poly/ML bug. *)
    (* *)
    fun add_axiom add_axioms_for_term t accum =
      let val t = unfold_defs_in_hol_term ext_ctx t in
        if member (op aconv) accum t then accum
        else add_axioms_for_term t (t :: accum)
      end
    (* *)
    fun add_named_axiom x name =
      case AList.lookup op= all_axioms name of
        SOME t => add_axiom add_axioms_for_term (Refute.specialize_type thy x t)
      | NONE => raise BadArg ("add_named_axiom", "unknown axiom " ^ quote name)
    (* *)
    and add_axioms_for_term t =
      case t of
        Const (x as (s, T)) =>
        (if is_built_in_const x then
           I
         else if s mem unsupported_consts then
           raise NotSupported ("axioms_for_hol_term", s)
         else
           case s of
             @{const_name The} => add_named_axiom x "HOL.the_eq_trivial"
           | @{const_name Hilbert_Choice.Eps} =>
             add_named_axiom x "Hilbert_Choice.someI"
           | _ =>
             if Refute.is_const_of_class thy x then
               let
                 val class = Logic.class_of_const s
                 val inclass = Logic.mk_inclass (TVar (("'a", 0), [class]),
                                                 class)
                 val ax1 = SOME (Refute.specialize_type thy x inclass)
                           handle Type.TYPE_MATCH => NONE
                 val ax2 = Option.map (Refute.specialize_type thy x o snd)
                                      (Refute.get_classdef thy class)
               in
                 fold (add_axiom add_axioms_for_term) (map_filter I [ax1, ax2])
               end
             else if is_equational_fun ext_ctx x then
               fold (add_axiom add_axioms_for_term)
                    (equational_fun_axioms ext_ctx x)
             else
               I)
        #> add_axioms_for_type T
      | Free (_, T) => add_axioms_for_type T
      | Var (_, T) => add_axioms_for_type T
      | Bound _ => I
      | Abs (_, T, t) => add_axioms_for_term t #> add_axioms_for_type T
      | t1 $ t2 => add_axioms_for_term t1 #> add_axioms_for_term t2
    (* *)
    and add_axioms_for_type T =
      case T of
        Type (s, Ts) =>
        fold add_axioms_for_type Ts
        #> (if s = @{const_name int}
               orelse isSome (DatatypePackage.get_datatype thy s) then
              I
            else
              add_axiom_for_typedef T)
      | TFree (_, S) => add_axioms_for_sort T S
      | TVar (_, S) => add_axioms_for_sort T S
    (* *)
    and add_axiom_for_typedef T =
      case Refute.get_typedef thy T of
        SOME (_, t) => add_axiom add_axioms_for_term t
      | NONE => I
    (* *)
    and add_axioms_for_sort T S =
      let
        val superclasses = S @ maps (Sign.super_classes thy) S |> distinct op=
        val class_axioms =
          maps (fn class =>
                   map Thm.prop_of (#axioms (AxClass.get_info thy class)
                                    handle ERROR _ => []))
               superclasses
        val monomorphic_class_axioms =
          map (fn t =>
                  case Term.add_tvars t [] of
                    [] => t
                  | [(j, S)] =>
                    Refute.monomorphic_term (Vartab.make [(j, (S, T))]) t
                  | _ => raise BadHOLTerm ("add_axioms_for_sort", [t]))
              class_axioms
      in fold (add_axiom add_axioms_for_term) monomorphic_class_axioms end
    val watch_axioms = map2 watch_axiom_for_hol_term
                            (0 upto length watch - 1) watch
  in
    add_axioms_for_term t []
    |> fold (add_axiom add_axioms_for_term) watch_axioms
    |> map (specialize_consts_in_hol_term ext_ctx true o Refute.close_form)
  end

(* (string * string list) list ref -> int -> bool -> hol_term -> hol_term *)
fun skolemize_hol_term skolems max_depth =
  let
    (* *)
    val incrs = map (curry op+ 1)
    (* *)
    fun aux ss Ts js depth neg t =
      let
        (* *)
        fun handle_quantifier quant_s quant_T abs_s abs_T t =
          if not (loose_bvar1 (t, 0)) then
            aux ss Ts js depth neg (incr_boundvars ~1 t)
          else if depth <= max_depth
                  andalso neg = (quant_s <> @{const_name Ex}) then
            let
              val j = length (!skolems) + 1
              val sko_s = skolem_prefix_for j ^ abs_s
              val _ = change skolems (cons (sko_s, ss))
              val sko_t = list_comb (Const (sko_s, rev Ts ---> abs_T),
                                     map Bound (rev js))
              val abs_t = Abs (abs_s, abs_T, aux ss Ts (incrs js) depth neg t)
            in
              if null js then betapply (abs_t, sko_t)
              else Const (@{const_name Let}, abs_T --> quant_T) $ sko_t $ abs_t
            end
          else
            Const (quant_s, quant_T)
            $ Abs (abs_s, abs_T,
                   (not (is_fun_type abs_T)
                    ? aux (abs_s :: ss) (abs_T :: Ts) (0 :: incrs js)
                          (depth + 1) neg) t)
      in
        case t of
          Const (s0 as @{const_name all}, T0) $ Abs (s1, T1, t1) =>
          handle_quantifier s0 T0 s1 T1 t1
        | Const (s0 as @{const_name All}, T0) $ Abs (s1, T1, t1) =>
          handle_quantifier s0 T0 s1 T1 t1
        | Const (s0 as @{const_name Ex}, T0) $ Abs (s1, T1, t1) =>
          handle_quantifier s0 T0 s1 T1 t1
        | @{const Trueprop} $ t1 =>
          @{const Trueprop} $ aux ss Ts js depth neg t1
        | @{const Not} $ t1 => @{const Not} $ aux ss Ts js depth (not neg) t1
        | @{const "==>"} $ t1 $ t2 =>
          @{const "==>"} $ aux ss Ts js depth (not neg) t1
          $ aux ss Ts js depth neg t2
        | @{const "op &"} $ t1 $ t2 =>
          @{const "op &"} $ aux ss Ts js depth neg t1
          $ aux ss Ts js depth neg t2
        | @{const "op |"} $ t1 $ t2 =>
          @{const "op |"} $ aux ss Ts js depth neg t1
          $ aux ss Ts js depth neg t2
        | @{const "op -->"} $ t1 $ t2 =>
          @{const "op -->"} $ aux ss Ts js depth (not neg) t1
          $ aux ss Ts js depth neg t2
        | (t0 as Const (@{const_name If}, _)) $ t1 $ t2 $ t3 =>
          t0 $ t1 $ aux ss Ts js depth neg t2 $ aux ss Ts js depth neg t3
        | (t0 as Const (@{const_name Let}, T0)) $ t1 $ t2 =>
          t0 $ t1 $ aux ss Ts js depth neg t2
        | Abs (s, T, t1) => Abs (s, T, aux ss Ts (incrs js) depth neg t1)
        | _ => t
      end
  in aux [] [] [] 0 end

(* extended_context -> bool -> hol_term -> hol_term list * hol_term *)
fun preprocess_hol_term (ext_ctx as {skolem_depth, skolems, ...}) negated =
  unfold_defs_in_hol_term ext_ctx
  #> skolemize_hol_term skolems skolem_depth negated
  #> specialize_consts_in_hol_term ext_ctx false
  #> `(axioms_for_hol_term ext_ctx)

(* Adapted from refute.ML *)
(* Proof.context -> hol_term list -> typ list *)
fun ground_types_in_hol_term ctx ts =
  let
    val thy = ProofContext.theory_of ctx
    (* typ -> typ list -> typ list *)
    fun add_type T =
      case T of
        Type ("fun", Ts) => fold add_type Ts
      | Type ("*", Ts) => fold add_type Ts
      | Type (@{type_name itself}, [T1]) => add_type T1
      | @{typ prop} => I
      | @{typ bool} => I
      | @{typ unit} => I
      | Type (s, Ts) =>
        (case DatatypePackage.get_datatype thy s of
           SOME {descr, index, ...} =>
           let
             val (_, typs, _) = the (AList.lookup op= descr index)
             val typ_assoc = typs ~~ Ts
             (* *)
             fun add_dtype U =
               let val T = Refute.typ_of_dtyp descr typ_assoc U in
                 case U of
                   DatatypeAux.DtTFree _ => add_type T
                 | DatatypeAux.DtType (_, Us) =>
                   add_type T #> fold add_dtype Us
                 | DatatypeAux.DtRec i =>
                   (fn accum =>
                       if T mem accum then
                         accum
                       else
                         let
                           val (_, dtyps, dconstrs) = the (AList.lookup op= descr i)
                           val accum = (not (is_pair_type T) ? insert op= T) accum
                                       |> fold add_dtype dtyps
                         in
                           fold add_dtype (List.concat (map snd dconstrs)) accum
                         end)
               end
           in add_dtype (DatatypeAux.DtRec index) end
         | NONE => insert op= T #> fold add_type Ts)
      | TFree _ => insert op= T
      | TVar _ => insert op= T
  in foldl (OldTerm.it_term_types (uncurry add_type)) [] ts end

end;
