(*  Title:      Nitpick/Tools/nitpick_isar.ML
    ID:         $Id: nitpick_isar.ML,v 1.81 2008/10/10 15:03:21 blanchet Exp $
    Author:     Jasmin Blanchette, TU Muenchen
    Copyright   2008, 2009

Adds the "nitpick" and "nitpick_params" commands to Isabelle/Isar's outer
syntax.
*)

structure NitpickIsar : sig end =
struct

open NitpickUtil
open NitpickRep
open NitpickTerm
open Nitpick

structure TheoryData = TheoryDataFun(
  type T = {params: (string * string) list}
  val empty = {params = []}
  val copy = I
  val extend = I
  fun merge _ ({params = ps1}, {params = ps2}) =
    {params = AList.merge op= op= (ps1, ps2)})

(* string * string -> theory -> theory *)
fun set_default_raw_param param thy =
  TheoryData.put {params = AList.update op= param
                               (#params (TheoryData.get thy))} thy
(* theory -> (string * string) list *)
val default_raw_params = #params o TheoryData.get

(* Proof.context -> (string * string) list -> params *)
fun extract_params ctx raw_params =
  let
    (* string -> string -> string *)
    fun lookup_string def_value name =
      getOpt (AList.lookup op= raw_params name, def_value)
    (* bool -> string -> bool *)
    fun lookup_bool def_value name =
      case AList.lookup op= raw_params name of
        SOME "true" => true
      | SOME "false" => false
      | SOME _ => error ("Parameter " ^ quote name ^
                         " must be assigned \"true\" or \"false\"")
      | NONE => def_value
    (* int -> string -> int *)
    fun lookup_int def_value name =
      case AList.lookup op= raw_params name of
        SOME s => (case Int.fromString s of
                     SOME i => i
                   | NONE => error ("Parameter " ^ quote name ^
                                    " must be assigned an integer value"))
      | NONE => def_value
    (* *)
    fun int_range_from_string s =
      (case space_explode "-" s of
         [s] => [the (Int.fromString s)]
       | [s1, s2] =>
         let val (k1, k2) = pairself (the o Int.fromString) (s1, s2) in
           if k1 <= k2 then k1 upto k2 else k1 downto k2
         end
       | _ => [])
      handle Option => []
    (* *)
    fun int_seq_from_string s =
      List.concat (map int_range_from_string (space_explode "," s))
    (* *)
    fun lookup_int_seq def_value name =
      case AList.lookup op= raw_params name of
        SOME s => (case int_seq_from_string s of
                     [] => def_value
                   | value => value)
      | NONE => def_value
    (* *)
    fun lookup_ints_assigns def_value prefix =
      ("", lookup_int_seq def_value prefix)
      :: map (fn (name, value) =>
                 (String.extract (name, size prefix + 1, NONE),
                  int_seq_from_string value))
             (filter (String.isPrefix (prefix ^ " ") o fst) raw_params)
    (* *)
    fun lookup_time name =
      case AList.lookup op= raw_params name of
        SOME "none" => Time.zeroTime
      | SOME s =>
        let
          val msecs =
            case space_explode " " s of
              [s1, "s"] => 1000 * the (Int.fromString s1)
            | [s1, "ms"] => the (Int.fromString s1)
            | _ => 0
        in
          if msecs <= 0 then
            error ("Parameter " ^ quote name ^ " must be assigned a time value \
                   \(e.g., \"60 s\", \"200 ms\", \"none\")")
          else
            Time.fromMilliseconds msecs
        end
      | NONE => Time.zeroTime
    (* *)
    fun lookup_mood def_value name =
      case AList.lookup op= raw_params name of
        SOME s => (case s of
                     "optimistic" => Optimistic
                   | "realistic" => Realistic
                   | "pessimistic" => Pessimistic
                   | _ => error ("Parameter " ^ quote name ^
                                 " must be assigned \"optimistic\", \
                                 \\"realistic\", or \"pessimistic\"."))
      | NONE => def_value
    (* *)
    fun lookup_hol_term_list def_value name =
      case AList.lookup op= raw_params name of
        SOME s => Syntax.read_terms ctx (space_explode ";" s)
      | NONE => def_value
    val cards_assigns = lookup_ints_assigns [1] "card"
    val mults_assigns = lookup_ints_assigns [1] "mult"
    val iters_assigns = lookup_ints_assigns [0] "iter"
    val lockstep = lookup_bool false "lockstep"
    val sat_solver = lookup_string "auto" "sat_solver"
    val batch_size = Int.max (1, lookup_int 1 "batch_size")
    val falsify = lookup_bool true "falsify"
    val normalize = lookup_bool false "normalize"
    val inductive_mood = lookup_mood Pessimistic "inductive_mood"
    val specialize_funs = lookup_bool false "specialize_funs"
    val skolem_depth = lookup_int 0 "skolem_depth"
    val peephole = lookup_bool false "peephole"
    val timeout = lookup_time "timeout"
    val auto_timeout = lookup_time "auto_timeout"
    val wellfounded_timeout = lookup_time "wellfounded_timeout"
    val sym_break = lookup_int 20 "sym_break"
    val sharing_depth = Int.max (1, lookup_int 3 "sharing_depth")
    val flatten_props = lookup_bool false "flatten_props"
    val verbose = lookup_bool false "verbose"
    val show_skolems = lookup_bool false "show_skolems"
    val show_datatypes = lookup_bool false "show_datatypes"
    val show_consts = lookup_bool false "show_consts"
    val watch = lookup_hol_term_list [] "watch"
    val max_potential = Int.max (0, lookup_int 0 "max_potential")
    val check_potential = lookup_bool false "check_potential"
    val check_genuine = lookup_bool false "check_genuine"
    val expect = lookup_string "" "expect"
  in
    {cards_assigns = cards_assigns, mults_assigns = mults_assigns,
     iters_assigns = iters_assigns, lockstep = lockstep, sat_solver = sat_solver,
     batch_size = batch_size, falsify = falsify, normalize = normalize,
     inductive_mood = inductive_mood, specialize_funs = specialize_funs,
     skolem_depth = skolem_depth, peephole = peephole, timeout = timeout,
     auto_timeout = auto_timeout, wellfounded_timeout = wellfounded_timeout,
     sym_break = sym_break, sharing_depth = sharing_depth,
     flatten_props = flatten_props, verbose = verbose,
     show_skolems = show_skolems, show_datatypes = show_datatypes,
     show_consts = show_consts, watch = watch, max_potential = max_potential,
     check_potential = check_potential, check_genuine = check_genuine,
     expect = expect}
  end

(* OuterParse.token list -> string * OuterParse.token list *)
val scan_key = Scan.repeat1 OuterParse.typ_group >> space_implode " "

(* *)
val scan_value =
  Scan.repeat1 (OuterParse.minus
                || Scan.repeat1 (Scan.unless OuterParse.minus OuterParse.name)
                   >> space_implode " "
                || OuterParse.$$$ "," |-- OuterParse.number >> prefix ",")
  >> foldr op^ ""

(* OuterParse.token list -> (bstring * bstring) * OuterParse.token list *)
val scan_param = scan_key -- (OuterParse.$$$ "=" |-- scan_value)
(* OuterParse.token list
   -> (bstring * bstring) list option * OuterParse.token list *)
val scan_params = Scan.option (OuterParse.$$$ "[" |-- OuterParse.list scan_param
                               --| OuterParse.$$$ "]")

(* *)
fun is_known_param s =
  s mem ["card", "mult", "iter", "lockstep", "sat_solver", "batch_size",
         "falsify", "normalize", "inductive_mood", "specialize_funs",
         "skolem_depth", "peephole", "timeout", "auto_timeout",
         "wellfounded_timeout", "sym_break", "sharing_depth", "flatten_props",
         "verbose", "show_skolems", "show_datatypes", "show_consts", "watch",
         "max_potential", "check_potential", "check_genuine", "expect"]
  orelse exists (fn p => String.isPrefix (p ^ " ") s) ["card", "mult", "iter"]
fun check_param (s, _) =
  if is_known_param s then () else error ("Unknown parameter " ^ quote s)

(* (TableFun().key * string) list option * int option -> Toplevel.transition
   -> Toplevel.transition *)
fun nitpick_trans (opt_params, opt_subgoal) =
  Toplevel.keep
      (fn state =>
          let
            val subgoal = getOpt (opt_subgoal, 1)
            val thm = snd (snd (Proof.get_goal (Toplevel.proof_of state)))
            val thy = Toplevel.theory_of state
            val ctx = Toplevel.context_of state
            val override_params = getOpt (opt_params, [])
            val _ = map check_param override_params
            val params = extract_params ctx (rev (default_raw_params thy @
                                                  override_params))
          in
            (satisfy_or_falsify_subgoal ctx params thm (subgoal - 1); ())
            handle BadArg (loc, details) =>
                   error ("Bad argument list to " ^ loc ^ ": " ^ details)
                 | BadHOLTerm (loc, ts) =>
                   error ("Invalid term" ^ plural_s ts ^ " in " ^ loc ^ ": " ^
                          commas (map (Syntax.string_of_term ctx) ts))
                 | BadRep (loc, Rs) =>
                   error ("Invalid representation" ^ plural_s Rs ^ " in " ^
                          loc ^ ": " ^ commas (map rep_to_string Rs))
                 | BadTerm (loc, us) =>
                   error ("Invalid term" ^ plural_s us  ^ " in " ^ loc ^ ": " ^
                          commas (map (term_to_string ctx) us))
                 | BadType (loc, Ts) =>
                   error ("Invalid type" ^ plural_s Ts ^ " in " ^ loc ^ ": " ^
                          commas (map (Syntax.string_of_typ ctx) Ts))
                 | Kodkod.BadSyntax (loc, details) =>
                   error ("Syntax error in " ^ loc ^ ": " ^ details)
                 | NotSupported (loc, details) =>
                   error ("Unsupported case in " ^ loc ^ ": " ^ details)
          end)

(* (TableFun().key * string) list option -> Toplevel.transition
   -> Toplevel.transition *)
fun nitpick_params_trans opt_params =
  Toplevel.theory
      (fn thy =>
          let
            val thy = fold set_default_raw_param (getOpt (opt_params, [])) thy
          in
            writeln ("Default parameters for Nitpick:\n" ^
                     (case rev (default_raw_params thy) of
                        [] => "none"
                      | params =>
                        (map check_param params;
                         cat_lines (map (fn (name, value) =>
                                            name ^ " = " ^ value) params))));
            thy
          end)

(* OuterParse.token list
   -> (Toplevel.transition -> Toplevel.transition) * OuterParse.token list *)
fun scan_nitpick_command tokens =
  (scan_params -- Scan.option OuterParse.nat) tokens |>> nitpick_trans
fun scan_nitpick_params_command tokens =
  scan_params tokens |>> nitpick_params_trans

val _ = OuterSyntax.improper_command "nitpick"
            "try to find a counterexample for a given subgoal using Kodkod"
            OuterKeyword.diag scan_nitpick_command
val _ = OuterSyntax.command "nitpick_params"
            "set and display the default parameters for Nitpick"
            OuterKeyword.thy_decl scan_nitpick_params_command

end;
