(*  Title:      Nitpick/Tools/nitpick_model.ML
    ID:         $Id: nitpick_model.ML,v 1.81 2008/01/20 21:49:51 blanchet Exp $
    Author:     Jasmin Blanchette, TU Muenchen
    Copyright   2009

Model reconstruction for Nitpick.
*)

signature NITPICK_MODEL =
sig
  type hol_term = Term.term
  type fun_special = NitpickHOL.fun_special
  type scope = NitpickScope.scope
  type rep = NitpickRep.rep
  type term = NitpickTerm.term

  type params = {
    show_skolems: bool,
    show_datatypes: bool,
    show_consts: bool}

  structure NameTable : TABLE

  val maybe_const_name : string ref
  val tuple_list_for_name :
    term NameTable.table -> Kodkod.raw_bound list -> term -> int list list
  val hol_term_for_name :
    bool -> scope -> term list -> term NameTable.table -> Kodkod.raw_bound list
    -> term -> hol_term
  val reconstruct_hol_model :
    params -> scope -> hol_term list -> (string * string list) list
    -> fun_special list -> term list -> term list -> term list
    -> term NameTable.table -> Kodkod.raw_bound list -> string
  val prove_hol_model :
    scope -> Time.time -> term list -> term list -> term NameTable.table
    -> Kodkod.raw_bound list -> bool -> hol_term -> bool option
end;

structure NitpickModel : NITPICK_MODEL =
struct

open NitpickUtil
open NitpickHOL
open NitpickScope
open NitpickPeephole
open NitpickRep
open NitpickTerm

type params = {
  show_skolems: bool,
  show_datatypes: bool,
  show_consts: bool}

val maybe_const_name = ref "" (* set in Nitpick.thy *)

(* *)
fun atom_name (T as Type (s, _)) j = nitpick_prefix ^ s ^ nat_subscript (j + 1)
  | atom_name (T as TFree (s, _)) j =
    (unprefix "'" s handle Fail _ => nitpick_prefix ^ s) ^ nat_subscript (j + 1)
  | atom_name T _ = raise BadType ("atom_name", [T])

(* term NameTable.table -> Kodkod.raw_bound list -> term -> int list list *)
fun tuple_list_for_name rel_table bounds name =
  the (AList.lookup op= bounds (lookup_rel rel_table name))
  handle BadTerm _ => [[]]

(* bool -> params -> scope -> term list -> term list -> term list
   -> term NameTable.table -> Kodkod.raw_bound list -> string *)
fun reconstruct_hol_term for_auto {card_assigns, datatypes, ...} ctor_names
                         rel_table bounds =
  let
    (* *)
    fun ctor_for_atom T j =
      the (find_first (fn {multiplicity, offset, ...} =>
                          j < offset + multiplicity)
                      (#ctors (the (datatype_spec datatypes T))))
    (* *)
    fun make_fun T1 T2 ts1 ts2 =
      let
        (* *)
        fun truth_const_sort_key @{const True} = "0"
          | truth_const_sort_key @{const False} = "2"
          | truth_const_sort_key _ = "1"
        (* *)
        fun aux_general [] [] =
            if for_auto andalso T2 = bool_T then Abs ("x", T1, @{const False})
            else Const (@{const_name undefined}, Type ("fun", [T1, T2]))
          | aux_general (t1 :: ts1) (t2 :: ts2) =
            Const (@{const_name fun_upd}, [T1 --> T2, T1, T2] ---> (T1 --> T2))
            $ aux_general ts1 ts2 $ t1 $ t2
          | aux_general _ _ = raise UnequalLengths
        (* *)
        fun aux_set [] = Const (@{const_name "{}"}, Type ("fun", [T1, T2]))
          | aux_set ((t1, t2) :: zs) =
            (t2 <> @{const False} ?
             curry op$ (Const (@{const_name insert},
                               [T1, T1 --> T2] ---> (T1 --> T2))
                        $ (t2 <> @{const True}
                           ? curry op$ (Const (!maybe_const_name, T1 --> T1)))
                          t1))
            (aux_set zs)
      in
        if for_auto orelse T2 <> bool_T then aux_general (rev ts1) (rev ts2)
        else aux_set (sort_wrt (truth_const_sort_key o snd) (ts1 ~~ ts2))
      end

    (* *)
    fun term_for_atom (T as Type ("fun", [T1, T2])) j =
        let
          val k1 = card_of_type card_assigns T1
          val k2 = card_of_type card_assigns T2
        in
          term_for_rep T (Vect (k1, Atom k2))
                       [nth_combination (replicate k1 k2) j]
        end
      | term_for_atom (Type ("*", [T1, T2])) j =
        let val k1 = card_of_type card_assigns T1 in
          list_comb (HOLogic.pair_const T1 T2,
                     map2 term_for_atom [T1, T2] [j div k1, j mod k1])
        end
      | term_for_atom @{typ bool} j =
        if j = 0 then @{const False} else @{const True}
      | term_for_atom @{typ unit} _ = @{const Unity}
      | term_for_atom T j =
        if T = nat_T then
          HOLogic.mk_number nat_T j
        else if T = int_T then
          HOLogic.mk_number int_T
                            (int_for_atom (card_of_type card_assigns int_T) j)
        else if is_iterator_type T then
          HOLogic.mk_number nat_T (card_of_type card_assigns T - j - 1)
        else if datatype_spec datatypes T = NONE then
          (if for_auto then Free else Const) (atom_name T j, T)
        else
          let
            val {name, typ, ...} = ctor_for_atom T j
            val R = the (get_first
                             (fn ConstName (s, T, R) =>
                                 if (s, T) = (name, typ) then SOME R else NONE
                               | u => raise BadTerm ("term_for_atom", [u]))
                             ctor_names)
            val jss = tuple_list_for_name rel_table bounds
                                          (ConstName (name, typ, R))
            val js = if R = Unit then
                       []
                     else if body_rep R = Formula then
                       (* ## correct? *)
                       if j = 1 then the_single jss
                       else hd (all_combinations_for_reps (binder_reps R))
                     else
                       the (get_first (fn js =>
                                          let val (js, j') = split_last js in
                                            if j' = j then SOME js else NONE
                                          end) jss)
            val Ts = binder_types typ
            val Rs = if R = Unit then replicate (length Ts) Unit
                     else binder_reps R
            val arities = map arity_of_rep Rs
            val jsss = map single (unevenly_chunk_list arities js)
            val ts = map3 term_for_rep Ts Rs jsss
          in list_comb (Const (name, typ), ts) end
   (* *)
    and term_for_vect k R T1 T2 js =
      make_fun T1 T2 (map (term_for_atom T1) (0 upto k - 1))
               (map (term_for_rep T2 R o single)
                    (batch_list (arity_of_rep R) js))
    (* *)
    and term_for_rep T Unit [[]] = term_for_rep T (Atom 1) [[0]]
      | term_for_rep T (R as Atom k) [[j]] =
        if j < k then term_for_atom T j else raise BadRep ("term_for_rep", [R])
      | term_for_rep (Type ("*", [T1, T2])) (Struct [R1, R2]) [js] =
        let
          val arity1 = arity_of_rep R1
          val (js1, js2) = chop arity1 js
        in
          list_comb (HOLogic.pair_const T1 T2,
                     map3 term_for_rep [T1, T2] [R1, R2] [[js1], [js2]])
        end
      | term_for_rep (Type ("fun", [T1, T2])) (R as Vect (k, R')) [js] =
        term_for_vect k R' T1 T2 js
      | term_for_rep (Type ("fun", [T1, T2])) (Func (R1, Formula)) jss =
        let
          val jss1 = all_combinations_for_rep R1
          val ts1 = map (term_for_rep T1 R1 o single) jss1
          val ts2 =
            map (fn js => term_for_rep T2 (Atom 2)
                                       [[bool_to_int (js mem jss)]]) jss1
        in make_fun T1 T2 ts1 ts2 end
      | term_for_rep (Type ("fun", [T1, T2])) (Func (R1, R2)) jss =
        let
          val arity1 = arity_of_rep R1
          val jss1 = all_combinations_for_rep R1
          val ts1 = map (term_for_rep T1 R1 o single) jss1
          val grouped_jss2 = AList.group op= (map (chop arity1) jss)
          val ts2 = map (term_for_rep T2 R2 o the_default []
                         o AList.lookup op= grouped_jss2) jss1
        in make_fun T1 T2 ts1 ts2 end
      | term_for_rep T (Opt R) jss =
        if null jss then Const ("\\<bottom>", T) else term_for_rep T R jss
      | term_for_rep T R jss =
        raise BadArg ("term_for_rep",
                      Refute.string_of_typ T ^ " " ^ rep_to_string R ^ " " ^
                      Int.toString (length jss))
  in term_for_rep end

(* bool -> scope -> term list -> term NameTable.table -> Kodkod.raw_bound list
   -> term -> hol_term *)
fun hol_term_for_name for_auto scope ctor_names rel_table bounds name =
  tuple_list_for_name rel_table bounds name
  |> reconstruct_hol_term for_auto scope ctor_names rel_table bounds
                          (type_of name) (rep_of name)

(* params -> scope -> hol_term list -> (string * string list) list
   -> fun_special list -> term list -> term list -> term list -> term list
   -> term NameTable.table -> string *)
fun reconstruct_hol_model {show_skolems, show_datatypes, show_consts}
        (scope as {ctx, card_assigns, datatypes}) watch skolems fun_specials
        free_names ctor_names non_ctor_names rel_table bounds =
  let
    val term_for_rep = reconstruct_hol_term false scope ctor_names rel_table
                                            bounds
    (* *)
    fun string_for_assign name =
      let
        val jss = tuple_list_for_name rel_table bounds name
        val t1 =
          case name of
            FreeName (s, T, _) => Free (s, T)
          | ConstName (s, T, _) =>
            user_friendly_const watch skolems fun_specials (s, T)
          | _ => raise BadTerm ("string_for_assign", [name])
        val t2 = term_for_rep (type_of name) (rep_of name) jss
      in Syntax.string_of_term ctx t1 ^ " = " ^ Syntax.string_of_term ctx t2 end
    (* *)
    fun string_for_datatype {typ, card, precise, ...} =
      Syntax.string_of_typ ctx typ ^ " = {" ^
      commas ((map (Syntax.string_of_term ctx)
                   (map (term_for_rep typ (Atom card) o single o single)
                        (0 upto card - 1)))
              @ (if precise then [] else ["\\<bottom>"])) ^ "}"
    (* *)
    fun pseudo_datatype T =
      case card_of_type card_assigns T of
        ~1 => []
      | k => [{typ = nat_T, card = k, precise = false, ctors = []}]
    val datatypes = pseudo_datatype nat_T @ pseudo_datatype int_T @ datatypes
    val block_of_datatypes =
      if show_datatypes andalso not (null datatypes) then
        "Datatype" ^ plural_s datatypes ^ ":\n" ^
        cat_lines (map (prefix "    " o string_for_datatype) datatypes) ^ "\n"
      else
        ""
    (* string -> term list -> string *)
    fun block_of_names show title names =
      if show andalso not (null names) then
        title ^ plural_s names ^ ":\n" ^
        cat_lines (map (prefix "    " o string_for_assign)
                       (sort_wrt nickname_of names)) ^ "\n"
      else
        ""
    val (skolem_names, non_skolem_non_ctor_names) =
      List.partition (String.isPrefix skolem_prefix o nickname_of)
                     non_ctor_names
    val (watch_names, real_const_names) =
      List.partition (String.isPrefix watch_prefix o nickname_of)
                     non_skolem_non_ctor_names
    val res = block_of_names true "Free variable" free_names ^
              block_of_names true "Watched term" watch_names ^
              block_of_names show_skolems "Skolem constant" skolem_names ^
              block_of_datatypes ^
              block_of_names show_consts "Constant" real_const_names
  in if res = "" then "Trivial" else unsuffix "\n" res end

(* scope -> Time.time -> term list -> term list -> term NameTable.table
  -> Kodkod.raw_bound list -> bool -> hol_term -> bool option *)
fun prove_hol_model (scope as {ctx, card_assigns, ...}) auto_timeout free_names
                    ctor_names rel_table bounds negate prop =
  let
    val thy = ProofContext.theory_of ctx
    (* *)
    fun free_type_assm (T, k) =
      let
        fun atom j = Free (atom_name T j, T)
        fun equation_for_atom j = HOLogic.eq_const T $ Bound 0 $ atom j
        fun inequation_for_atoms i j =
          @{term Not} $ HOLogic.mk_eq (atom i, atom j)
        fun inequations_for_atom i =
          map (inequation_for_atoms i) (i + 1 upto k - 1)
        val eqs = map equation_for_atom (0 upto k - 1)
(*
        val compreh_assm =
          Const (@{const_name All}, (T --> bool_T) --> bool_T)
              $ Abs ("x", T, foldl1 HOLogic.mk_disj eqs)
*)
        val distinct_assms =
          map (foldl1 HOLogic.mk_conj o inequations_for_atom) (0 upto k - 2)
      in foldl (HOLogic.mk_conj o swap) @{const True} distinct_assms end

    (* *)
    fun free_name_assm name =
      HOLogic.mk_eq (Free (nickname_of name, type_of name),
                     hol_term_for_name true scope ctor_names rel_table bounds
                                       name)

    val freeT_assms = map free_type_assm (filter (is_TFree o fst) card_assigns)
    val model_assms = map free_name_assm free_names
    val assm = foldr HOLogic.mk_conj @{const True} (freeT_assms @ model_assms)
    (* *)
    fun try_out negate =
      let
        val concl = (negate ? curry op$ @{term Not})
                    (ObjectLogic.atomize_term thy prop)
        val goal = HOLogic.mk_Trueprop (HOLogic.mk_imp (assm, concl))
                   |> map_types (map_type_tfree
                          (fn (s, []) => TFree (s, HOLogic.typeS)
                            | x => TFree x))
                   |> cterm_of thy |> Goal.init
      in
        (goal |> SINGLE (timeout_tac auto_timeout (CLASIMPSET auto_tac))
              |> the |> Goal.finish; true)
        handle THM _ => false
             | TimeLimit.TimeOut => false
      end
  in
    if silence try_out negate then SOME true
    else if silence try_out (not negate) then SOME false
    else NONE
  end

end;
