(*  Title:      Nitpick/Tools/nitpick_scope.ML
    ID:         $Id: nitpick_scope.ML,v 1.81 2008/12/12 09:13:33 blanchet Exp $
    Author:     Jasmin Blanchette, TU Muenchen
    Copyright   2008, 2009

Scope enumerator for Nitpick.
*)

signature NITPICK_SCOPE =
sig
  type ctor_spec = {
    name: string,
    typ: typ,
    multiplicity: int,
    total: bool,
    self_rec: bool,
    offset: int}

  type dtype_spec = {
    typ: typ,
    card: int,
    precise: bool,
    ctors: ctor_spec list}

  type scope = {
    ctx: Proof.context,
    card_assigns: (typ * int) list,
    datatypes: dtype_spec list}

  val datatype_spec : dtype_spec list -> typ -> dtype_spec option
  val ctor_spec : dtype_spec list -> string * typ -> ctor_spec
  val is_typedef : theory -> string -> bool
  val is_datatype_like : theory -> typ -> bool
  val precise_card_of_type : theory -> (typ * int) list -> typ -> int
  val card_of_type : (typ * int) list -> typ -> int
  val is_precise_type : dtype_spec list -> typ -> bool
  val is_empty_scope : scope -> bool
  val compact_string_for_scope : scope -> string
  val spacious_string_for_scope : scope -> string
  val all_scopes :
    Proof.context -> (string * int list) list -> (string * int list) list
    -> (string * int list) list -> typ list -> bool -> scope list
end;

structure NitpickScope : NITPICK_SCOPE =
struct

open NitpickUtil
open NitpickHOL

type ctor_spec = {
  name: string,
  typ: typ,
  multiplicity: int,
  total: bool,
  self_rec: bool,
  offset: int}

type dtype_spec = {
  typ: typ,
  card: int,
  precise: bool,
  ctors: ctor_spec list}

type scope = {
  ctx: Proof.context,
  card_assigns: (typ * int) list,
  datatypes: dtype_spec list}

datatype entry =
  Typ of typ |
  Ctor of string * typ
type row = entry * int list
type block = row list

(* dtype_spec list -> typ -> dtype_spec option *)
fun datatype_spec dtypes T = List.find (equal T o #typ) dtypes

(* dtype_spec list -> string * typ -> ctor_spec *)
fun ctor_spec [] z = raise BadHOLTerm ("ctor_spec", [Const z])
  | ctor_spec ({ctors, ...} :: dtypes) z =
    (case List.find (equal z o (#name pairf #typ)) ctors of
       SOME c => c
     | NONE => ctor_spec dtypes z)

(* theory -> string -> bool *)
val is_typedef = isSome oo TypedefPackage.get_info
(* theory -> typ -> bool *)
fun is_datatype_like thy (T as Type (s, _)) =
    (isSome (DatatypePackage.get_datatype thy s) orelse is_typedef thy s
     orelse T = @{typ ind})
    andalso not (is_integer_type T)
  | is_datatype_like _ _ = false

(* *)
fun datatype_ctors thy (T as Type (s, Ts)) =
    (case DatatypePackage.get_datatype thy s of
       SOME {descr, index, ...} =>
       let val (_, dtyps, ctors) = the (AList.lookup op= descr index) in
         map (fn (name, Us) =>
                 (name, map (Refute.typ_of_dtyp descr (dtyps ~~ Ts)) Us ---> T))
             (rev ctors)
       end     
     | NONE =>
       case TypedefPackage.get_info thy s of
         SOME {abs_type, rep_type, Abs_name, ...} =>
         [(Abs_name, instantiate_type thy abs_type T rep_type --> T)]
       | NONE =>
         if T = @{typ ind} then
           [dest_Const @{const Suc_Rep}, dest_Const @{const Zero_Rep}]
         else
           raise BadType ("datatype_ctors", [T]))
  | datatype_ctors _ T = raise BadType ("datatype_ctors", [T])

(* *)
fun precise_card_of_type thy asgns =
  let
    fun aux avoid T =
      if T mem avoid then
        0
      else
        case T of
          Type ("fun", [T1, T2]) =>
            (case aux avoid T1 of 0 => 0 | k1 => power (aux avoid T2) k1)
        | Type (@{type_name itself}, _) => 1
        | @{typ prop} => 2
        | @{typ bool} => 2
        | @{typ unit} => 1
        | Type (s, Ts) =>
          if is_datatype_like thy T then
            let
              val ctor_cards = datatype_ctors thy T
                               |> map (binder_types o snd)
                               |> map (foldl op* 1 o map (aux (T :: avoid)))
            in
              if exists (equal 0) ctor_cards then 0 else foldl op+ 0 ctor_cards
            end
          else if is_integer_type T then
            0
          else
            the_default 0 (AList.lookup op= asgns T)
        | _ => the_default 0 (AList.lookup op= asgns T)
  in aux [] end

(* *)
fun card_of_type asgns (Type ("fun", [T1, T2])) =
    power (card_of_type asgns T2) (card_of_type asgns T1)
  | card_of_type asgns (Type ("*", [T1, T2])) =
    card_of_type asgns T1 * card_of_type asgns T2
  | card_of_type _ (Type (@{type_name itself}, _)) = 2
  | card_of_type _ @{typ prop} = 2
  | card_of_type _ @{typ bool} = 2
  | card_of_type _ @{typ unit} = 1
  | card_of_type asgns T =
    case AList.lookup op= asgns T of
      SOME k => k
    | NONE => if is_integer_type T then ~1
              else raise BadType ("card_of_type", [T])

(* dtype_spec list -> typ -> bool *)
fun is_precise_type dtypes (Type ("fun", Ts)) =
    forall (is_precise_type dtypes) Ts
  | is_precise_type dtypes (Type ("*", Ts)) = forall (is_precise_type dtypes) Ts
  | is_precise_type dtypes T =
    not (is_integer_type T) andalso #precise (the (datatype_spec dtypes T))
    handle Option => true

(* scope -> string * string * string *)
fun string_triple_for_scope {ctx, card_assigns, datatypes} =
  let
    val thy = ProofContext.theory_of ctx
    val (iter_assigns, proper_card_assigns) =
      List.partition (is_iterator_type o fst)
                     (filter (not o is_datatype_like thy o fst) card_assigns)
    val cards = map (fn (T, k) => type_to_string ctx T ^ " = " ^ Int.toString k)
                    proper_card_assigns
    val mults = List.concat (map (map (fn {name, multiplicity, ...} =>
                                          name ^ " = " ^
                                          Int.toString multiplicity)
                                  o #ctors) datatypes)
    val iters = map (fn (T, k) =>
                        type_to_string ctx T ^ " = " ^ Int.toString (k - 1))
                    iter_assigns
  in (cards, mults, iters) end

(* scope -> bool *)
fun is_empty_scope {card_assigns, datatypes, ...} =
  null card_assigns andalso null datatypes

(* scope -> string *)
fun compact_string_for_scope scope =
  let
    val (cards, mults, iters) = string_triple_for_scope scope
    val ss = map (prefix "card ") cards @ map (prefix "mult ") mults @
             map (prefix "iter ") iters
  in if null ss then "\\<emptyset>" else serial_commas ss end

(* scope -> string *)
fun spacious_string_for_scope scope =
  let val (cards, mults, iters) = string_triple_for_scope scope in
    case (if null cards then [] else ["card: " ^ commas cards]) @
         (if null mults then [] else ["mult: " ^ commas mults]) @ 
         (if null iters then [] else ["iter: " ^ commas iters]) of
      [] => "empty"
    | lines => space_implode "\n" lines
  end

(* row -> int *)
fun rank_of_row (_, ks) = length ks
(* block -> int *)
fun rank_of_block block = foldl Int.max 1 (map rank_of_row block)
(* *)
fun project_row column (y, ks) =
  (y, [nth ks (Int.min (column, length ks - 1))])
(* *)
fun project_block column = map (project_row column)

(* *)
fun block_for_plain_type ctx cards_asgns T =
  [(Typ T, map (curry Int.max 1)
               (case AList.lookup op= cards_asgns (type_to_string ctx T) of
                  SOME cards => cards
                | NONE =>
                  case AList.lookup op= cards_asgns "" of
                    SOME cards => cards
                  | NONE => raise BadArg ("block_for_plain_type", "")))]

(* ### improve *)
(* *)
fun score_of_ctor_selector (name, T) stor =
  if stor = "" then
    0
  else
    case space_explode "." name of
      [twig] => if stor = twig then 7 else ~1
    | [thy, twig] =>
      if String.isPrefix "Abs_" twig then
        let val twiglie = String.extract (twig, 4, NONE) in
          case space_explode "." stor of
            [s1] => if s1 = twig then 4
                    else if s1 = twiglie then 3
                    else ~1
          | [s1, s2] => if (s1, s2) = (thy, twig) then 6
                        else if (s1, s2) = (thy, twiglie) then 5
                        else ~1
          | _ => raise BadArg ("score_of_ctor_selector",
                               "selector " ^ quote stor)
        end
      else
        raise BadArg ("score_of_ctor_selector", "ctor " ^ quote name)
    | [thy, typ, twig] =>
      (case space_explode "." stor of
         [s1] => if s1 = twig then 4
                 else if s1 = typ then 2
                 else if s1 = thy then 1
                 else ~1
       | [s1, s2] => if (s1, s2) = (typ, twig) then 6
                     else if (s1, s2) = (thy, twig) then 5
                     else if (s1, s2) = (thy, typ) then 3
                     else ~1
       | [_, _, _] => if name = stor then 7 else ~1
       | _ => raise BadArg ("score_of_ctor_selector", "selector " ^ quote stor)) 
    | _ => raise BadArg ("score_of_ctor_selector", "ctor " ^ quote name)

(* *)
fun row_for_ctor mults_asgns ctor =
  let
    val scored_asgns = map (apfst (score_of_ctor_selector ctor)) mults_asgns
    val top_score = foldl1 Int.max (map fst scored_asgns)
  in (Ctor ctor, the (AList.lookup op= scored_asgns top_score)) end

(* *)
fun block_for_datatype ctx cards_asgns mults_asgns iters_asgns T =
  if is_iterator_type T then
    block_for_plain_type ctx (map (apsnd (map (curry op+ 1))) iters_asgns) T
  else if is_integer_type T then
    block_for_plain_type ctx cards_asgns T
  else
    map (row_for_ctor mults_asgns)
        (datatype_ctors (ProofContext.theory_of ctx) T)

(* *)
fun block_for_datatypes ctx cards_asgns mults_asgns iters_asgns Ts =
  List.concat
      (map (block_for_datatype ctx cards_asgns mults_asgns iters_asgns) Ts)

(* *)
fun blocks_for_types ctx cards_asgns mults_asgns iters_asgns Ts lockstep =
  let
    val thy = ProofContext.theory_of ctx
    val (dataTs, free_Ts) =
      List.partition (is_integer_type orf is_datatype_like thy) Ts
    val free_T_blocks = map (block_for_plain_type ctx cards_asgns) free_Ts
    val dataT_block = block_for_datatypes ctx cards_asgns mults_asgns
                                          iters_asgns dataTs
    val blocks = dataT_block :: free_T_blocks
  in if lockstep then [List.concat blocks] else blocks end

(* int list -> int list list *)
fun all_combinations_ordered_judiciously [] = [[]]
  | all_combinations_ordered_judiciously (special_k :: ks) =
    let
      fun combinations_for [] _ = [[]]
        | combinations_for [k] sum = if sum < k then [[sum]] else []
        | combinations_for (k :: ks) sum =
          List.concat (List.tabulate (Int.min (k, sum + 1),
              fn j => map (cons j) (combinations_for ks (sum - j))))
      fun max_sum ks = foldl op+ (~(length ks)) ks
      val jss = List.concat (map (combinations_for ks) (0 upto max_sum ks))
    in map_product cons (0 upto special_k - 1) jss end

(* *)
val ctor_type_of_multiplicity_assign = snd o fst
(* *)
val datatype_of_multiplicity_assign =
  body_type o ctor_type_of_multiplicity_assign

(* *)
fun datatype_card_assign_from_multiplicity_assigns asgns T =
  (T, map snd (filter (equal T o datatype_of_multiplicity_assign) asgns)
      |> foldl op+ 0)

(* *)
fun is_consistent_multiplicity_assign card_asgns ((_, T), mult) =
  let
    val dom_card = foldl op* 1 (map (card_of_type card_asgns) (binder_types T))
  in mult <= dom_card end
  handle BadType _ => true

(* *)
fun all_card_assigns card_asgns pending_mult_asgns mult_asgns =
  let
    val pending_Ts =
      sort_distinct TermOrd.typ_ord
                    (map datatype_of_multiplicity_assign pending_mult_asgns)
    val known_Ts = map datatype_of_multiplicity_assign mult_asgns
                   |> sort_distinct TermOrd.typ_ord
    val fully_known_Ts = OrdList.subtract TermOrd.typ_ord pending_Ts known_Ts
  in
    card_asgns @ map (datatype_card_assign_from_multiplicity_assigns mult_asgns)
                     fully_known_Ts
  end

(* *)
fun is_consistent_multiplicity_assign_set card_asgns pending_asgns asgns =
  forall (is_consistent_multiplicity_assign
              (all_card_assigns card_asgns pending_asgns asgns)) asgns

(* *)
fun is_self_recursive_ctor_type T =
  let val (binder_Ts, body_T) = strip_type T in
    exists (exists_subtype (equal body_T)) binder_Ts
  end

(* *)
fun is_total_ctor card_asgns self_rec mult T =
  let val binder_Ts = binder_types T in
    not self_rec
    andalso mult = foldl op* 1 (map (card_of_type card_asgns) binder_Ts)
  end

(* *)
fun ctor_spec_from_multiplicity_assign card_asgns self_rec offset
                                       ((name, T), mult) =
  {name = name, typ = T, multiplicity = mult,
   total = is_total_ctor card_asgns self_rec mult T, self_rec = self_rec,
   offset = offset}

(* *)
fun datatype_spec_from_multiplicity_assigns thy card_asgns T asgns =
  let
    val offsets = offset_list (map snd asgns)
    val self_recs = map (is_self_recursive_ctor_type
                         o ctor_type_of_multiplicity_assign) asgns
    val ctors = map3 (ctor_spec_from_multiplicity_assign card_asgns)
                     self_recs offsets asgns
    val card = card_of_type card_asgns T
    val precise = (card > 0
                   andalso card = precise_card_of_type thy card_asgns T)
  in {typ = T, card = card, precise = precise, ctors = ctors} end

(* *)
fun datatype_specs_from_multiplicity_assigns thy card_asgns asgns =
  let
    val asgns_by_Ts = map (fn a as ((_, T), _) => (body_type T, a)) asgns
                      |> AList.coalesce op=
    val all_card_asgns = all_card_assigns card_asgns [] asgns
  in
    map (uncurry (datatype_spec_from_multiplicity_assigns thy all_card_asgns))
        asgns_by_Ts
  end

(* *)
fun fix_multiplicity_assigns card_asgns mult_asgns =
  let
    fun aux seen [] = SOME seen
      | aux seen ((_, ~1) :: _) = NONE
      | aux seen ((y, mult) :: asgns) =
        if is_consistent_multiplicity_assign_set card_asgns asgns
                                                 ((y, mult) :: seen) then
          case aux ((y, mult) :: seen) asgns of
            SOME assigns => SOME assigns
          | NONE => aux seen ((y, mult - 1) :: asgns)
        else
          aux seen ((y, mult - 1) :: asgns)
  in
    the (aux [] mult_asgns)
    handle Option => raise BadArg ("fix_multiplicity_assigns",
                                   "unresolvable inconsistencies")
  end

(* *)
fun fix_iterator_assigns card_asgns mult_asgns =
  let
    val all_asgns = all_card_assigns card_asgns [] mult_asgns
    (* *)
    fun fix_one (T as Type (s, (T' :: _)), k) =
        (T, (is_iterator_type T ? curry Int.min (card_of_type all_asgns T')) k)
      | fix_one z = z
  in map fix_one card_asgns end

(* *)
fun partition_row (Typ T, [card]) (card_asgns, mult_asgns) =
    ((T, card) :: card_asgns, mult_asgns)
  | partition_row (Ctor (s, T), [mult]) (card_asgns, mult_asgns) =
    (card_asgns, ((s, T), mult) :: mult_asgns)
  | partition_row _ _ = raise BadArg ("partition_row", "")

(* *)
fun scope_descriptor_from_block block = fold_rev partition_row block ([], [])

(* block list -> int list -> scope *)
fun scope_descriptor_from_combination blocks columns =
  let
    val block = List.concat (map2 project_block columns blocks)
    val (card_asgns, mult_asgns) = scope_descriptor_from_block block
    val mult_asgns = fix_multiplicity_assigns card_asgns mult_asgns
    val card_asgns = fix_iterator_assigns card_asgns mult_asgns
  in (card_asgns, mult_asgns) end

(* *)
fun scope_from_descriptor ctx (card_asgns, mult_asgns) =
  {ctx = ctx, card_assigns = all_card_assigns card_asgns [] mult_asgns,
   datatypes = datatype_specs_from_multiplicity_assigns
                   (ProofContext.theory_of ctx) card_asgns mult_asgns}

(* Proof.context -> (string * int list) list -> (string * int list) list
   -> (string * int list) list -> typ list -> bool -> scope list *)
fun all_scopes ctx cards_asgns mults_asgns iters_asgns Ts lockstep =
  let
    val blocks = blocks_for_types ctx cards_asgns mults_asgns iters_asgns Ts
                                  lockstep
    val ranks = map rank_of_block blocks
    val descs = map (scope_descriptor_from_combination blocks)
                    (all_combinations_ordered_judiciously ranks)
  in map (scope_from_descriptor ctx) (distinct op= descs) end

end;
