(*  Title:      Nitpick/Tools/nitpick_util.ML
    ID:         $Id: nitpick_util.ML,v 1.81 2008/12/12 09:13:33 blanchet Exp $
    Author:     Jasmin Blanchette, TU Muenchen
    Copyright   2008, 2009

General-purpose functions used by the Nitpick modules.
*)

infix 6 nat_minus
infix 7 pairf

signature NITPICK_UTIL =
sig
  type hol_term = Term.term

  exception BadArg of string * string
  exception BadHOLTerm of string * hol_term list
  exception BadType of string * typ list
  exception Limit of string * string
  exception NotSupported of string * string
  exception Same

  val nitpick_prefix : string
  val curry3 : ('a * 'b * 'c -> 'd) -> 'a -> 'b -> 'c -> 'd
  val pairf : ('a -> 'b) * ('a -> 'c) -> 'a -> 'b * 'c
  val bool_to_int : bool -> int
  val nat_minus : int * int -> int
  val power : int -> int -> int
  val exact_log : int -> int -> int
  val exact_root : int -> int -> int
  val fold1 : ('a -> 'a -> 'a) -> 'a list -> 'a
  val replicate_list : int -> 'a list -> 'a list
  val n_fold_cartesian_product : 'a list list -> 'a list list
  val nth_combination : int list -> int -> int list
  val all_combinations : int list -> int list list
  val batch_list : int -> 'a list -> 'a list list
  val unevenly_chunk_list : int list -> 'a list -> 'a list list
  val map3 : ('a -> 'b -> 'c -> 'd) -> 'a list -> 'b list -> 'c list -> 'd list
  val serial_commas : string list -> string
  val plural_s : 'a list -> string
  val offset_list : int list -> int list
  val index_seq : int -> int -> int list
  val filter_indices : int list -> 'a list -> 'a list
  val filter_out_indices : int list -> 'a list -> 'a list
  val bool_T : typ
  val nat_T : typ
  val int_T : typ
  val subscript : string -> string
  val nat_subscript : int -> string
  val time_limit : Time.time -> ('a -> 'b) -> 'a -> 'b
  val silence : ('a -> 'b) -> 'a -> 'b
  val timeout_tac : Time.time -> tactic -> tactic
end

structure NitpickUtil : NITPICK_UTIL =
struct

type hol_term = Term.term

exception BadArg of string * string
exception BadHOLTerm of string * hol_term list
exception BadType of string * typ list
exception Limit of string * string
exception NotSupported of string * string
exception Same

val nitpick_prefix = "Nitpick."

(* ('a * 'b * 'c -> 'd) -> 'a -> 'b -> 'c -> 'd *)
fun curry3 f = fn x => fn y => fn z => f (x, y, z)

(* ('a -> 'b) * ('a -> 'c) -> 'a -> 'b * 'c *)
fun (f pairf g) = fn x => (f x, g x)

(* bool -> int *)
fun bool_to_int b = if b then 1 else 0
(* int * int -> int *)
fun (i nat_minus j) = if i > j then i - j else 0

(* int -> int -> int *)
fun power a 0 = 1
  | power a 1 = a
  | power a b = let val c = power a (b div 2) in c * c * power a (b mod 2) end

(* int -> int -> int *)
fun exact_log m n =
  let
    val res = Real.round (Math.ln (Real.fromInt n) / Math.ln (Real.fromInt m))
  in
    if power m res = n then res
    else raise BadArg ("exact_log", Int.toString m ^ " " ^ Int.toString n)
  end

(* int -> int -> int *)
fun exact_root m n =
  let val res = Real.round (Math.pow (Real.fromInt n, 1.0 / (Real.fromInt m))) in
    if power res m = n then res
    else raise BadArg ("exact_root", Int.toString m ^ " " ^ Int.toString n)
  end

(* ('a -> 'a -> 'a) -> 'a list -> 'a *)
fun fold1 f = foldl1 (uncurry f)

(* int -> 'a list -> 'a list *)
fun replicate_list 0 _ = []
  | replicate_list n xs =
    if n mod 2 = 0 then replicate_list (n div 2) (xs @ xs)
    else xs @ replicate_list (n - 1) xs

(* 'a list list -> 'a list list *)
fun n_fold_cartesian_product xss = fold_rev multiply xss [[]]

(* int list -> int -> int list *)
val nth_combination =
  let
    fun aux [] n = ([], n)
      | aux (m :: ms) n =
        let val (js, n) = aux ms n in ((n mod m) :: js, n div m) end
  in fst oo aux end

(* int list -> int list list *)
val all_combinations = n_fold_cartesian_product o map (fn k => 0 upto k - 1)

(* int -> 'a list -> 'a list list *)
fun batch_list k xs =
  if length xs <= k then [xs]
  else List.take (xs, k) :: batch_list k (List.drop (xs, k))

(* int list -> 'a list -> 'a list list *)
fun unevenly_chunk_list [] [] = []
  | unevenly_chunk_list (k :: ks) ys =
    let val (ys1, ys2) = chop k ys in ys1 :: unevenly_chunk_list ks ys2 end
  | unevenly_chunk_list _ _ = raise BadArg ("unevenly_chunk_list", "")

(* ('a -> 'b -> 'c -> 'd) -> 'a list -> 'b list -> 'c list -> 'd list *)
fun map3 _ [] [] [] = []
  | map3 f (x :: xs) (y :: ys) (z :: zs) = f x y z :: map3 f xs ys zs
  | map3 _ _ _ _ = raise UnequalLengths

(* string list -> string *)
fun serial_commas [] = "???"
  | serial_commas [s] = s
  | serial_commas [s1, s2] = s1 ^ " and " ^ s2
  | serial_commas ss = commas (List.take (ss, length ss - 1)) ^ ", and " ^
                       List.last ss

(* 'a list -> string *)
fun plural_s xs = if length xs = 1 then "" else "s"

(* int list -> int list *)
val offset_list = rev o tl o foldl (fn (x, xs) => (x + hd xs) :: xs) [0]

(* int -> int -> int list *)
fun index_seq j0 n = if j0 < 0 then j0 downto j0 - n + 1 else j0 upto j0 + n - 1

(* int list -> 'a list -> 'a list *)
fun filter_indices js xs =
  let
    fun aux _ [] _ = []
      | aux i (j :: js) (x :: xs) =
        if i = j then x :: aux (i + 1) js xs else aux (i + 1) (j :: js) xs
      | aux _ _ _ = raise BadArg ("filter_indices",
                                  "indices unordered or out of range")
  in aux 0 js xs end
fun filter_out_indices js xs =
  let
    fun aux _ [] xs = xs
      | aux i (j :: js) (x :: xs) =
        if i = j then aux (i + 1) js xs else x :: aux (i + 1) (j :: js) xs
      | aux _ _ _ = raise BadArg ("filter_out_indices",
                                  "indices unordered or out of range")
  in aux 0 js xs end

val bool_T = @{typ bool}
val nat_T = @{typ nat}
val int_T = @{typ int}

(* string -> string *)
val subscript = implode o map (prefix "\\<^isub>") o explode
(* int -> string *)
val nat_subscript = subscript o Int.toString

(* Time.time -> ('a -> 'b) -> 'a -> 'b *)
fun time_limit delay f x =
  (if delay = Time.zeroTime then f else TimeLimit.timeLimit delay f) x

(* (string -> unit) ref -> ('a -> 'b) -> 'a -> 'b *)
fun silence_one out_fn f x =
  let
    val old_out_fn = !out_fn
    val _ = out_fn := K ()
    val y = f x
    val _ = out_fn := old_out_fn
  in y end

(* ('a -> 'b) -> 'a -> 'b *)
fun silence f =
  fold silence_one
       [Output.writeln_fn, Output.priority_fn, Output.tracing_fn,
        Output.warning_fn, Output.error_fn, Output.debug_fn, Output.prompt_fn,
        Output.status_fn] f

(* Time.time -> tactic -> tactic *)
fun timeout_tac delay tac state =
  case time_limit delay (tac #> Seq.pull) state of
    SOME (x, _) => Seq.single x
  | NONE => Seq.empty

end;
